"""Console script for oceantide."""
import sys
import click


@click.command()
def main(args=None):
    """Console script for oceantide."""
    click.echo("Replace this message by putting your code into " "oceantide.cli.main")
    click.echo("See click documentation at http://click.pocoo.org/")
    return 0



"""Test Otis functions."""
import filecmp
from pathlib import Path

from oceantide import read_otis_netcdf
from oceantide.core.otis import (
    read_otis_bin_h,
    read_otis_bin_u,
    read_otis_bin_grid,
    write_otis_bin_h,
    write_otis_bin_u,
    write_otis_bin_grid,
)


FILES_DIR = Path(__file__).parent / "test_files"


def test_read_otis_netcdf():
    hfile = FILES_DIR / "otis_netcdf/hf.test.nc"
    ufile = FILES_DIR / "otis_netcdf/uv.test.nc"
    gfile = FILES_DIR / "otis_netcdf/grid.test.nc"
    dset = read_otis_netcdf(gfile=gfile, hfile=hfile, ufile=ufile)
    assert hasattr(dset, "tide")


def test_binary_otis_io(tmpdir):
    """Read existing binaries, write new binaries, compare existing and new files."""

    hfile0 = FILES_DIR / "otis_binary/h_rag"
    ufile0 = FILES_DIR / "otis_binary/u_rag"
    gfile0 = FILES_DIR / "otis_binary/grid_rag"

    hfile1 = tmpdir / "h"
    ufile1 = tmpdir / "u"
    gfile1 = tmpdir / "g"

    # Reading
    dsh = read_otis_bin_h(hfile0)
    dsu = read_otis_bin_u(ufile0)
    dsg = read_otis_bin_grid(gfile0)

    # Writing
    write_otis_bin_h(hfile1, dsh.hRe, dsh.hIm, dsh.con, dsh.lon_z, dsh.lat_z)
    write_otis_bin_u(
        ufile1, dsu.URe, dsu.UIm, dsu.VRe, dsu.VIm, dsu.con, dsh.lon_z, dsh.lat_z
    )
    write_otis_bin_grid(gfile1, dsg.hz, dsg.mz, dsh.lon_z, dsh.lat_z, dt=12)

    assert filecmp.cmp(hfile0, hfile1, shallow=False)
    assert filecmp.cmp(ufile0, ufile1, shallow=False)

    # Grid files are different due to wrong iob in original files but values are set when reading
    dsg1 = read_otis_bin_grid(gfile1)
    assert dsg.identical(dsg1)


def test_otis_binary_writer_plugin(tmpdir):
    "Test output otis binary plugin."
    dset = read_otis_netcdf(FILES_DIR / "otis_netcdf/Model_test")
    filenames = dset.tide.to_otis_binary(tmpdir, hfile=True, ufile=True, gfile=True)
    print(filenames)



"""Test tide accessor."""
from pathlib import Path
import pytest
import datetime
import xarray as xr

from oceantide import read_oceantide


FILES_DIR = Path(__file__).parent / "test_files"


@pytest.fixture(scope="module")
def dset():
    filename = FILES_DIR / "oceantide.zarr"
    _dset = read_oceantide(filename)
    yield _dset


def test_accessor_created(dset):
    assert hasattr(dset, "tide")
    assert hasattr(dset.tide, "predict")


def test_predict_elevation_only(dset):
    times = [datetime.datetime(2001, 1, 1, H) for H in range(6)]
    eta = dset.tide.predict(times, components=["h"])
    assert "u" not in eta and "v" not in eta and "h" in eta


def test_predict_current_only(dset):
    times = [datetime.datetime(2001, 1, 1, H) for H in range(6)]
    eta = dset.tide.predict(times, components=["u", "v"])
    assert "h" not in eta and "u" in eta and "v" in eta


def test_always_predict_something(dset):
    times = [datetime.datetime(2001, 1, 1, H) for H in range(6)]
    with pytest.raises(ValueError):
        dset.tide.predict(times, components=[])



"""Otis binary output."""
from pathlib import Path
import numpy as np
import xarray as xr

from oceantide.core.otis import u_from_z, v_from_z, indices_open_boundary
from oceantide.core.utils import set_attributes


def to_otis_netcdf(
    self,
    dirname: str,
    hfile: bool = True,
    ufile: bool = True,
    gfile: bool = True,
    suffix: bool = None,
) -> dict:
    """Write dataset as Otis binary format.

    Parameters
    ----------
    self (oceantide.tide.Tide)
        Oceantide Tide instance.
    dirname (str)
        Directory to save binary files.
    hfile (bool)
        Save tidal elevation binary file.
    ufile (bool)
        Save tidal transports binary file.
    gfile (bool)
        Save grid file.
    suffix (str)
        Suffix to define file names, by default defined by cons names.

    Returns
    -------
    filename (dict)
        Name of files written.

    """
    ds = self._obj.transpose("con", "lon", "lat", ...)
    ds = ds.rename({"con": "nc", "lon": "nx", "lat": "ny"})
    ds = ds.fillna(0.0)

    if suffix is None:
        suffix = "".join(list(ds.nc.values)).lower()

    lat_z, lon_z = np.meshgrid(ds.ny, ds.nx)
    dx = float(ds.nx[1] - ds.nx[0])
    dy = float(ds.ny[1] - ds.ny[0])
    ds = ds.drop(["nc", "nx", "ny"])
    cons = np.array([f"{c:4s}".lower() for c in self._obj.con.values], dtype="S4")

    mz = xr.where(ds.dep > 0, 1.0, 0.0)
    if ufile or gfile:
        mu = mz * mz.roll(nx=1, roll_coords=False)
        mv = mz * mz.roll(ny=1, roll_coords=False)
        lon_u = lon_z - dx
        lat_u = lat_z
        lon_v = lon_z
        lat_v = lat_z - dy

    filenames = {}

    # Write elevations
    if hfile:
        fname = f"h_{suffix.lstrip('_')}.nc" if suffix else "h.nc"
        filenames["hfile"] = Path(dirname) / fname

        ha = self.amplitude("h").transpose("con", "lon", "lat", ...)
        ha = ha.rename({"con": "nc", "lon": "nx", "lat": "ny"}).drop(["nc", "nx", "ny"])
        hp = self.phase("h").transpose("con", "lon", "lat", ...)
        hp = hp.rename({"con": "nc", "lon": "nx", "lat": "ny"}).drop(["nc", "nx", "ny"])
        dsout = xr.Dataset()
        dsout["con"] = xr.DataArray(cons, dims=("nc"))
        dsout["lon_z"] = xr.DataArray(lon_z, dims=("nx", "ny"))
        dsout["lat_z"] = xr.DataArray(lat_z, dims=("nx", "ny"))
        dsout["ha"] = xr.DataArray(ha, dims=("nc", "nx", "ny"))
        dsout["hp"] = xr.DataArray(hp, dims=("nc", "nx", "ny"))
        dsout["hRe"] = xr.DataArray(ds.h.real, dims=("nc", "nx", "ny"))
        dsout["hIm"] = xr.DataArray(ds.h.imag, dims=("nc", "nx", "ny"))

        set_attributes(dsout, "otis")
        dsout.attrs = {"title": "OTIS tidal elevation file", "source": "Oceantide"}

        dsout.to_netcdf(filenames["hfile"])

    # Write transports
    if ufile:
        fname = f"u_{suffix.lstrip('_')}.nc" if suffix else "u.nc"
        filenames["ufile"] = Path(dirname) / fname

        ua = self.amplitude("u").transpose("con", "lon", "lat", ...)
        ua = ua.rename({"con": "nc", "lon": "nx", "lat": "ny"}).drop(["nc", "nx", "ny"])
        up = self.phase("u").transpose("con", "lon", "lat", ...)
        up = up.rename({"con": "nc", "lon": "nx", "lat": "ny"}).drop(["nc", "nx", "ny"])
        va = self.amplitude("v").transpose("con", "lon", "lat", ...)
        va = va.rename({"con": "nc", "lon": "nx", "lat": "ny"}).drop(["nc", "nx", "ny"])
        vp = self.phase("v").transpose("con", "lon", "lat", ...)
        vp = vp.rename({"con": "nc", "lon": "nx", "lat": "ny"}).drop(["nc", "nx", "ny"])
        dsout = xr.Dataset()
        dsout["con"] = dsout["con"] = xr.DataArray(cons, dims=("nc"))
        dsout["lon_u"] = xr.DataArray(lon_u, dims=("nx", "ny"))
        dsout["lat_u"] = xr.DataArray(lat_u, dims=("nx", "ny"))
        dsout["lon_v"] = xr.DataArray(lon_v, dims=("nx", "ny"))
        dsout["lat_v"] = xr.DataArray(lat_v, dims=("nx", "ny"))
        dsout["ua"] = u_from_z(xr.DataArray(ua, dims=("nc", "nx", "ny")), mz, mu)
        dsout["up"] = u_from_z(xr.DataArray(up, dims=("nc", "nx", "ny")), mz, mu)
        dsout["va"] = v_from_z(xr.DataArray(va, dims=("nc", "nx", "ny")), mz, mv)
        dsout["vp"] = v_from_z(xr.DataArray(vp, dims=("nc", "nx", "ny")), mz, mv)
        dsout["URe"] = u_from_z(
            xr.DataArray(ds.u.real * ds.dep, dims=("nc", "nx", "ny")), mz, mu
        )
        dsout["UIm"] = u_from_z(
            xr.DataArray(ds.u.imag * ds.dep, dims=("nc", "nx", "ny")), mz, mu
        )
        dsout["VRe"] = v_from_z(
            xr.DataArray(ds.v.real * ds.dep, dims=("nc", "nx", "ny")), mz, mv
        )
        dsout["VIm"] = v_from_z(
            xr.DataArray(ds.v.imag * ds.dep, dims=("nc", "nx", "ny")), mz, mv
        )

        set_attributes(dsout, "otis")
        dsout.attrs = {
            "title": "OTIS tidal transport/current file",
            "source": "Oceantide",
        }

        dsout.to_netcdf(filenames["ufile"])

    # Write grid
    if gfile:
        fname = f"grid_{suffix.lstrip('_')}.nc" if suffix else "grid.nc"
        filenames["gfile"] = Path(dirname) / fname

        dsout = xr.Dataset()
        dsout["lon_z"] = xr.DataArray(lon_z, dims=("nx", "ny"))
        dsout["lat_z"] = xr.DataArray(lat_z, dims=("nx", "ny"))
        dsout["lon_u"] = xr.DataArray(lon_u, dims=("nx", "ny"))
        dsout["lat_u"] = xr.DataArray(lat_u, dims=("nx", "ny"))
        dsout["lon_v"] = xr.DataArray(lon_v, dims=("nx", "ny"))
        dsout["lat_v"] = xr.DataArray(lat_v, dims=("nx", "ny"))
        dsout["mz"] = xr.DataArray(mz, dims=("nx", "ny"))
        dsout["mu"] = xr.DataArray(mu, dims=("nx", "ny"))
        dsout["mv"] = xr.DataArray(mv, dims=("nx", "ny"))
        dsout["hz"] = xr.DataArray(ds.dep, dims=("nx", "ny"))
        dsout["hu"] = u_from_z(xr.DataArray(ds.dep, dims=("nx", "ny")), mz, mu)
        dsout["hv"] = v_from_z(xr.DataArray(ds.dep, dims=("nx", "ny")), mz, mv)
        dsout["iob_z"] = xr.DataArray(
            indices_open_boundary(dsout.mz), dims=("iiob", "nob_z")
        )
        dsout["iob_u"] = xr.DataArray(
            indices_open_boundary(dsout.mu), dims=("iiob", "nob_u")
        )
        dsout["iob_v"] = xr.DataArray(
            indices_open_boundary(dsout.mv), dims=("iiob", "nob_v")
        )

        set_attributes(dsout, "otis")
        dsout.attrs = {"title": "OTIS Arakawa C-grid file", "source": "Oceantide"}

        dsout.to_netcdf(filenames["gfile"])

    fname = f"model_{suffix.lstrip('_')}" if suffix else "model"
    with open(Path(dirname) / fname, mode="w") as stream:
        for filename in filenames.values():
            stream.write(f"./{filename.name}\n")

    return filenames


if __name__ == "__main__":
    from oceantide import read_otis_netcdf

    dset = read_otis_netcdf("/data/tide/tpxo9v4a/netcdf/DATA/Model_tpxo9v4a")
    dset = dset.isel(lon=slice(None, None, 10), lat=slice(None, None, 10)).load()
    filenames = dset.tide.to_otis_netcdf("./", hfile=True, ufile=True, gfile=True)
    print(filenames)



"""Top-level package for oceantide."""

__version__ = "0.7.2"


def _import_read_functions(pkgname: str = "input"):
    """Make read functions available at module level.

    Parameters
    ----------
    pkgname (str)
        Name of the package to import functions from.

    Functions are imported here if:

    - they are defined in a module wavespectra.input.{modname}
    - they are named as read_{modname}

    """
    import os
    import glob
    from importlib import import_module

    here = os.path.dirname(os.path.abspath(__file__))
    for filename in glob.glob1(os.path.join(here, pkgname), "*.py"):
        module = os.path.splitext(filename)[0]
        if module == "__init__":
            continue
        func_name = f"read_{module}"
        try:
            globals()[func_name] = getattr(
                import_module(f"oceantide.{pkgname}.{module}"), func_name
            )
        except Exception as exc:
            print(f"Cannot import reading function {func_name}:\n{exc}")


_import_read_functions()



"""Read Otis binary file format."""
import os
import xarray as xr

from oceantide.core.otis import (
    read_otis_bin_grid,
    read_otis_bin_h,
    read_otis_bin_u,
    otis_to_oceantide,
    otis_filenames,
)
from oceantide.tide import Tide


def read_otis_binary(
    filename: str = None, gfile: str = None, hfile: str = None, ufile: str = None
):
    """Read Otis Binary file format.

    Parameters
    ----------
    filename (str)
        Name of `Model_*` metadata file specifying other files to read.
    gfile (str)
        Name of grid file to read, by default defined from `filename`.
    hfile (str)
        Name of elevation file to read, by default defined from `filename`.
    ufile (str)
        Name of currents file to read, by default defined from `filename`.

    Returns
    -------
    dset (xr.Dataset)
        Formatted dataset with the tide accessor.

    Notes
    -----
    Otis data are usually provided in 4 separate files:

    - Meta file named as `Model_*` specifyig the 3 data files to read.
    - Grid file named as `grid*` with grid information and model depths.
    - Elevation file named as `h*` with elevation constituents data.
    - Transport file named as `uv` with transport constituents data.

    The path of the three data files can be prescribed either by specifying the meta
    file path (`filename`) or by explicitly providing their path (`gfile`, `hfile` and
    `ufile`).

    """
    if filename is not None:
        _gfile, _hfile, _ufile = otis_filenames(filename)
    else:
        _gfile = _hfile = _ufile = None
        if not all([gfile, hfile, ufile]):
            raise ValueError(
                "Either specify `filename` or all of `gfile`, `hfile`, `ufile`."
            )

    gfile = gfile or _gfile
    hfile = hfile or _hfile
    ufile = ufile or _ufile

    dsg = read_otis_bin_grid(gfile)
    dsh = read_otis_bin_h(hfile)
    dsu = read_otis_bin_u(ufile)

    dset = otis_to_oceantide(dsg, dsh, dsu)

    return dset


if __name__ == "__main__":
    hfile = "/data/tide/tpxo9v4a/bin/DATA/h_tpxo9.v4a"
    ufile = "/data/tide/tpxo9v4a/bin/DATA/u_tpxo9.v4a"
    gfile = "/data/tide/tpxo9v4a/bin/DATA/grid_tpxo9.v4a"

    dset = read_otis_binary(gfile=gfile, hfile=hfile, ufile=ufile)



"""Test tide accessor."""
from pathlib import Path
import pytest
import datetime
import xarray as xr

from oceantide import read_otis_netcdf, read_otis_binary, read_oceantide


FILES_DIR = Path(__file__).parent / "test_files"


def test_read_otis_binary_with_filename():
    read_otis_binary(FILES_DIR / "otis_binary/Model_rag")


def test_read_otis_binary_without_filename():
    read_otis_binary(
        gfile=FILES_DIR / "otis_binary/grid_rag",
        hfile=FILES_DIR / "otis_binary/h_rag",
        ufile=FILES_DIR / "otis_binary/u_rag",
    )


def test_otis_binary_correct_args():
    with pytest.raises(ValueError):
        read_otis_netcdf(
            hfile=FILES_DIR / "otis_binary/h_rag",
            ufile=FILES_DIR / "otis_binary/u_rag",
        )


def test_read_otis_netcdf_with_filename():
    read_otis_netcdf(FILES_DIR / "otis_netcdf/Model_test")


def test_read_otis_netcdf_without_filename():
    read_otis_netcdf(
        gfile=FILES_DIR / "otis_netcdf/grid.test.nc",
        hfile=FILES_DIR / "otis_netcdf/hf.test.nc",
        ufile=FILES_DIR / "otis_netcdf/uv.test.nc",
    )


def test_otis_netcdf_correct_args():
    with pytest.raises(ValueError):
        read_otis_netcdf(
            hfile=FILES_DIR / "otis_netcdf/hf.test.nc",
            ufile=FILES_DIR / "otis_netcdf/uv.test.nc",
        )


def test_read_write_oceantide_netcdf(tmpdir):
    dset = read_oceantide(FILES_DIR / "oceantide.nc")
    dset.tide.to_oceantide(tmpdir / "newoceantide.nc")


def test_read_write_oceantide_zarr(tmpdir):
    dset = read_oceantide(FILES_DIR / "oceantide.zarr")
    dset.tide.to_oceantide(str(tmpdir / "newoceantide.zarr"))


def test_supported_oceantide_formats(tmpdir):
    dset = read_oceantide(FILES_DIR / "oceantide.zarr")
    with pytest.raises(ValueError):
        dset.tide.to_oceantide(tmpdir / "newoceantide.zarr", file_format="txt")
    with pytest.raises(ValueError):
        dset.tide.to_oceantide(tmpdir / "newoceantide.txt")
    dset.tide.to_oceantide(tmpdir / "newoceantide.txt", file_format="netcdf")


def test_write_otis_netcdf(tmpdir):
    dset = read_otis_netcdf(FILES_DIR / "otis_netcdf/Model_test")
    dset.tide.to_otis_netcdf(dirname=tmpdir, suffix="")
    dset2 = read_otis_netcdf(tmpdir / "model")


def test_write_otis_binary(tmpdir):
    dset = read_otis_binary(FILES_DIR / "otis_binary/Model_rag")
    dset.tide.to_otis_binary(dirname=tmpdir, suffix="")
    dset2 = read_otis_binary(tmpdir / "model")



"""Otis core functions."""
import os
import re
import numpy as np
import dask.array as da
import xarray as xr

from oceantide.core.utils import arakawa_grid, set_attributes


CHAR = np.dtype(">c")
INT = np.dtype(">i4")
FLOAT = np.dtype(">f4")


def theta_lim(lon: np.ndarray, lat: np.ndarray) -> np.ndarray:
    """Grid limits (y0, y1, x0, x1).

    Parameters
    ----------
    lon (DataArray, 1darray)
        Longitude coordinates at the cell centre (Z-nodes).
    lat (DataArray, 1darray)
        Latitude coordinates at the cell centre (Z-nodes).

    Returns
    -------
    theta_lim (1darray)
        Grid limits with dtype for writing to binary file.

    """
    if lon.ndim > 1:
        if np.sum(np.diff(lon[:, 0])) != 0:
            lon = lon[:, 0]
        else:
            lon = lon[0, :]
    if lat.ndim > 1:
        if np.sum(np.diff(lat[0, :])) != 0:
            lat = lat[0, :]
        else:
            lat = lat[:, 0]
    dx = lon[1] - lon[0]
    dy = lat[1] - lat[0]
    if dx == 0 or dy == 0:
        raise ValueError(f"Longitude or latitude values are all equal")
    x0 = float(lon[0] - dx / 2)
    x1 = float(lon[-1] + dx / 2)
    y0 = float(lat[0] - dy / 2)
    y1 = float(lat[-1] + dy / 2)
    return np.hstack([y0, y1, x0, x1]).astype(FLOAT)


def indices_open_boundary(mz: np.ndarray):
    """Indices of open boundary iob.

    Parameters
    ----------
    mz (DataArray)
        Land mask in Otis format :math:`m_z(nx,ny)`.

    Returns
    -------
    iob (2darray)
        Indices of open boundary :math:`iob(2,nob)`. Longitude and latitude indices
        are defined in the first and second rows respectively.

    """
    nx, ny = mz.shape

    # Define indices coordinates
    mz = mz.assign_coords({"nx": range(1, nx + 1), "ny": range(1, ny + 1)}).fillna(1.0)

    # Slice four boundaries, sort so they are continuous, stack along new coord
    dsl = mz.isel(ny=[0]).stack({"n": ["nx", "ny"]})
    dst = mz.isel(nx=[-1]).stack({"n": ["nx", "ny"]})
    dsr = mz.isel(ny=[-1]).sortby("nx", ascending=False).stack({"n": ["nx", "ny"]})
    dsb = mz.isel(nx=[0]).sortby("ny", ascending=False).stack({"n": ["nx", "ny"]})

    # Concatenate boundaries and keep water points
    iob = xr.concat([dsl, dst, dsr, dsb], dim="n")
    iob = iob.where((iob == 1).compute(), drop=True)

    return np.array([iob.nx.values, iob.ny.values], dtype=INT)


def u_from_z(
    dz: xr.DataArray, mz: xr.DataArray = None, mu: xr.DataArray = None
) -> xr.DataArray:
    """Interpolate data from Z-nodes onto U-Nodes.

    Parameters
    ----------
    dz (DataArray)
        Data at Z-nodes to interpolate :math:`d_z(...,nx,ny)`.
    mz (DataArray)
        Land mask at Z-nodes :math:`m_z(nx,ny)`.
    mu (DataArray)
        Land mask at U-nodes :math:`m_u(nx,ny)`.

    Returns
    -------
    du (DataArray)
        Data interpolated at U-nodes :math:`d_u(...,nx,ny)`.

    Notes
    -----

    - Masks should be defined by 0 over land and either 1 or nan over water.
    - If mz is not provided it is defined from missing values in dz.
    - If mu is not provided it is interpolated from mz.

    """
    if mz is None:
        mz = xr.where(dz.isnull(), 0.0, 1.0)
    if mu is None:
        mu = mz * mz.roll(nx=1, roll_coords=False)
    iku = (mz.fillna(1.0) + mz.fillna(1.0).roll(nx=1, roll_coords=False)) == 1
    if "nc" in dz.dims:
        iku = iku.expand_dims({"nc": dz.nc})
    du = (dz + dz.roll(nx=1, roll_coords=False)) * mu.fillna(1.0) / 2
    du = xr.where(iku, dz, du)
    return du


def v_from_z(
    dz: xr.DataArray, mz: xr.DataArray = None, mv: xr.DataArray = None
) -> xr.DataArray:
    """Interpolate data from Z-nodes onto V-Nodes.

    Parameters
    ----------
    dz (DataArray)
        Data at Z-nodes to interpolate :math:`d_z(...,nx,ny)`.
    mz (DataArray)
        Land mask at Z-nodes :math:`m_z(nx,ny)`.
    mv (DataArray)
        Land mask at V-nodes :math:`m_v(nx,ny)`.

    Returns
    -------
    dv (DataArray)
        Data interpolated at V-nodes :math:`d_v(...,nx,ny)`.

    Notes
    -----

    - Masks should be defined by 0 over land and either 1 or nan over water.
    - If mz is not provided it is defined from missing values in dz.
    - If mv is not provided it is interpolated from mz.

    """
    if mz is None:
        mz = xr.where(dz.isnull(), 0.0, 1.0)
    if mv is None:
        mv = mz * mz.roll(ny=1, roll_coords=False)
    ikv = (mz.fillna(1.0) + mz.fillna(1.0).roll(ny=1, roll_coords=False)) == 1
    if "nc" in dz.dims:
        ikv = ikv.expand_dims({"nc": dz.nc})
    dv = (dz + dz.roll(ny=1, roll_coords=False)) * mv.fillna(1.0) / 2
    dv = xr.where(ikv, dz, dv)
    return dv


def z_from_u(
    du: xr.DataArray, mu: xr.DataArray = None, mz: xr.DataArray = None
) -> xr.DataArray:
    """Interpolate data from U-nodes onto Z-nodes.

    Parameters
    ----------
    du (DataArray)
        Data at U-nodes to interpolate :math:`d_u(...,nx,ny)`.
    mu (DataArray)
        Land mask at U-nodes :math:`m_u(nx,ny)`.
    mz (DataArray)
        Land mask at Z-nodes :math:`m_z(nx,ny)`.

    Returns
    -------
    dz (DataArray)
        Data interpolated at Z-nodes :math:`d_z(...,nx,ny)`.

    """
    if mu is None:
        mu = xr.where(du.isnull(), 0.0, 1.0)
    if mz is None:
        mz = mu * mu.roll(nx=-1, roll_coords=False)
    ikz = (mu.fillna(1.0) + mu.fillna(1.0).roll(nx=-1, roll_coords=False)) == 1
    if "nc" in du.dims:
        ikz = ikz.expand_dims({"nc": du.nc})
    dz = (du + du.roll(nx=-1, roll_coords=False)) * mz.fillna(1.0) / 2
    dz = xr.where(ikz, du, dz)
    return dz


def z_from_v(dv, mv=None, mz=None):
    """Interpolate data from V-nodes onto Z-nodes.

    Args:
        - dv (DataArray): Data at V-nodes to interpolate :math:`d_v(...,nx,ny)`.
        - mv (DataArray): Land mask at V-nodes :math:`m_v(nx,ny)`.
        - mz (DataArray): Land mask at Z-nodes :math:`m_z(nx,ny)`.

    Returns:
        - dz (DataArray): Data interpolated at Z-nodes :math:`d_z(...,nx,ny)`.

    """
    if mv is None:
        mv = xr.where(dv.isnull(), 0.0, 1.0)
    if mz is None:
        mz = mv * mv.roll(ny=-1, roll_coords=False)
    ikz = (mv.fillna(1.0) + mv.fillna(1.0).roll(ny=-1, roll_coords=False)) == 1
    if "nc" in dv.dims:
        ikz = ikz.expand_dims({"nc": dv.nc})
    dz = (dv + dv.roll(ny=-1, roll_coords=False)) * mz.fillna(1.0) / 2
    dz = xr.where(ikz, dv, dz)
    return dz


def otis_filenames(filename):
    """Otis data file names from `Model_*` metadata file.

    Args:
        filename (str): Name of `Model_*` metadata file specifying other files to read.

    Returns:
        gfile (str): Name of grid file to read, by default defined from `filename`.
        hfile (str): Name of elevation file to read, by default defined from `filename`.
        ufile (str): Name of currents file to read, by default defined from `filename`.

    """
    with open(filename) as stream:
        files = stream.read().split()
    dirname = os.path.dirname(filename)
    gfile = ufile = hfile = None
    for f in files:
        if os.path.split(f)[-1].lower().startswith("g"):
            gfile = os.path.join(dirname, os.path.basename(f))
        elif os.path.split(f)[-1].lower().startswith("u"):
            ufile = os.path.join(dirname, os.path.basename(f))
        elif os.path.split(f)[-1].lower().startswith("h"):
            hfile = os.path.join(dirname, os.path.basename(f))
    assert gfile, f"Cannot identify gfile from {filename} ({files})"
    assert hfile, f"Cannot identify hfile from {filename} ({files})"
    assert ufile, f"Cannot identify ufile from {filename} ({files})"
    return gfile, hfile, ufile


def read_otis_bin_u(ufile):
    """Read transport constituents data from otis binary file.

    Args:
        - ufile (str): Name of transport constituents binary file to read.

    Returns:
        - dset (Dataset): Transport constituents grid with variables:
            - URe: Real component of eastward transport :math:`U_{Re}(nc,nx,ny)`.
            - UIm: Imag component of eastward transport :math:`U_{Im}(nc,nx,ny)`.
            - VRe: Real component of northward transport :math:`V_{Re}(nc,nx,ny)`.
            - VIm: Imag component of northward transport :math:`V_{Im}(nc,nx,ny)`.

    """
    with open(ufile, "rb") as f:
        ll, nx, ny, nc = np.fromfile(f, dtype=INT, count=4)
        y0, y1, x0, x1 = np.fromfile(f, dtype=FLOAT, count=4)
        cons = [np.fromfile(f, CHAR, 4).tobytes().upper() for i in range(nc)]

    URe = np.zeros((nc, ny, nx))
    UIm = np.zeros((nc, ny, nx))
    VRe = np.zeros((nc, ny, nx))
    VIm = np.zeros((nc, ny, nx))

    for ic in range(nc):
        with open(ufile, "rb") as f:
            np.fromfile(f, dtype=INT, count=4)
            np.fromfile(f, dtype=FLOAT, count=4)

            nskip = int((ic) * (nx * ny * 16 + 8) + 8 + ll - 28)
            f.seek(nskip, 1)
            data = np.fromfile(f, dtype=FLOAT, count=4 * nx * ny).reshape((ny, 4 * nx))

        URe[ic] = data[:, 0 : 4 * nx - 3 : 4]
        UIm[ic] = data[:, 1 : 4 * nx - 2 : 4]
        VRe[ic] = data[:, 2 : 4 * nx - 1 : 4]
        VIm[ic] = data[:, 3 : 4 * nx - 0 : 4]

    URe = URe.transpose((0, 2, 1))
    UIm = UIm.transpose((0, 2, 1))
    VRe = VRe.transpose((0, 2, 1))
    VIm = VIm.transpose((0, 2, 1))

    # Dataset Otis style
    cons = np.array([c.ljust(4).lower() for c in cons])
    lon_u, lat_u = arakawa_grid(nx, ny, x0, x1, y0, y1, "u")
    lon_v, lat_v = arakawa_grid(nx, ny, x0, x1, y0, y1, "v")
    lat_u, lon_u = np.meshgrid(lat_u, lon_u)
    lat_v, lon_v = np.meshgrid(lat_v, lon_v)

    dset = xr.Dataset()
    dset["con"] = xr.DataArray(da.from_array(cons), dims=("nc",))
    dset["lon_u"] = xr.DataArray(da.from_array(lon_u), dims=("nx", "ny"))
    dset["lat_u"] = xr.DataArray(da.from_array(lat_u), dims=("nx", "ny"))
    dset["lon_v"] = xr.DataArray(da.from_array(lon_v), dims=("nx", "ny"))
    dset["lat_v"] = xr.DataArray(da.from_array(lat_v), dims=("nx", "ny"))
    dset["URe"] = xr.DataArray(da.from_array(URe), dims=("nc", "nx", "ny"))
    dset["UIm"] = xr.DataArray(da.from_array(UIm), dims=("nc", "nx", "ny"))
    dset["VRe"] = xr.DataArray(da.from_array(VRe), dims=("nc", "nx", "ny"))
    dset["VIm"] = xr.DataArray(da.from_array(VIm), dims=("nc", "nx", "ny"))

    # Attributes
    set_attributes(dset, "otis")
    dset.attrs = {
        "type": "OTIS tidal transport file",
        "title": "Oceantide tidal transport from binary file",
    }

    return dset


def read_otis_bin_h(hfile):
    """Read elevation constituents data from otis binary file.

    Args:
        - hfile (str): Name of elevation constituents binary file to read.

    Returns:
        - dset (Dataset): Elevation constituents grid with variables:
            - hRe: Real component of tidal elevation :math:`h_{Re}(nc,nx,ny)`.
            - hIm: Imag component of tidal elevation :math:`h_{Im}(nc,nx,ny)`.

    """
    with open(hfile, "rb") as f:
        ll, nx, ny, nc = np.fromfile(f, dtype=INT, count=4)
        y0, y1, x0, x1 = np.fromfile(f, dtype=FLOAT, count=4)
        cons = [np.fromfile(f, CHAR, 4).tobytes().upper() for i in range(nc)]

    hRe = np.zeros((nc, ny, nx))
    hIm = np.zeros((nc, ny, nx))

    for ic in range(nc):
        with open(hfile, "rb") as f:
            np.fromfile(f, dtype=INT, count=4)
            np.fromfile(f, dtype=FLOAT, count=4)

            nskip = int((ic) * (nx * ny * 8 + 8) + 8 + ll - 28)
            f.seek(nskip, 1)

            data = np.fromfile(f, dtype=FLOAT, count=2 * nx * ny).reshape((ny, 2 * nx))
            hRe[ic] = data[:, 0 : 2 * nx - 1 : 2]
            hIm[ic] = data[:, 1 : 2 * nx : 2]

    hRe = hRe.transpose((0, 2, 1))
    hIm = hIm.transpose((0, 2, 1))

    # Dataset Otis style
    lon_z, lat_z = arakawa_grid(nx, ny, x0, x1, y0, y1, "h")
    lat_z, lon_z = np.meshgrid(lat_z, lon_z)
    cons = np.array([c.ljust(4).lower() for c in cons])

    dset = xr.Dataset()
    dset["con"] = xr.DataArray(da.from_array(cons), dims=("nc",))
    dset["lon_z"] = xr.DataArray(da.from_array(lon_z), dims=("nx", "ny"))
    dset["lat_z"] = xr.DataArray(da.from_array(lat_z), dims=("nx", "ny"))
    dset["hRe"] = xr.DataArray(da.from_array(hRe), dims=("nc", "nx", "ny"))
    dset["hIm"] = xr.DataArray(da.from_array(hIm), dims=("nc", "nx", "ny"))

    # Attributes
    set_attributes(dset, "otis")
    dset.attrs = {
        "type": "OTIS tidal elevation file",
        "title": "Oceantide tidal elevation from binary file",
    }

    return dset


def read_otis_bin_grid(gfile):
    """Read grid data from otis binary file.

    Args:
        - gfile (str): Name of grid binary file to read.

    Returns:
        - dset (Dataset): Grid with variables:
            - lon_z: Longitude coordinates at Z-nodes :math:`lon_z(nx,ny)`.
            - lat_z: Latitude coordinates at Z-nodes :math:`lat_z(nx,ny)`.
            - lon_u: Longitude coordinates at U-nodes :math:`lon_u(nx,ny)`.
            - lat_u: Latitude coordinates at U-nodes :math:`lat_u(nx,ny)`.
            - lon_v: Longitude coordinates at V-nodes :math:`lon_v(nx,ny)`.
            - lat_v: Latitude coordinates at V-nodes :math:`lat_v(nx,ny)`.
            - mz: Mask at Z-nodes :math:`m_z(nx,ny)`.
            - mu: Mask at U-nodes :math:`m_u(nx,ny)`.
            - mv: Mask at V-nodes :math:`m_v(nx,ny)`.
            - hz: Depth at Z-nodes :math:`h_z(nx,ny)`.
            - hu: Depth at U-nodes :math:`h_u(nx,ny)`.
            - hv: Depth at V-nodes :math:`h_v(nx,ny)`.
            - iob_z: Indices of open boundary in hz :math:`iob_z(2,nob)`.
            - iob_u: Indices of open boundary in hu :math:`iob_u(2,nob)`.
            - iob_v: Indices of open boundary in hv :math:`iob_v(2,nob)`.

    Note:
        - Indices of open boundary are defined from mz, mu, mv, the values of iob
          read from file are ignored (there are inconsistencies in tpxo files).

    """
    with open(gfile, "rb") as f:
        f.seek(4, 0)
        nx, ny = np.fromfile(f, dtype=INT, count=2)
        y0, y1, x0, x1 = np.fromfile(f, dtype=FLOAT, count=4)
        dt = np.fromfile(f, dtype=FLOAT, count=1)  # lat-lon if > 0
        nob = np.fromfile(f, dtype=INT, count=1)[0]
        if nob == 0:
            f.seek(20, 1)
            iob = []
        else:
            f.seek(8, 1)
            iob = np.fromfile(f, INT, count=int(2 * nob)).reshape((2, int(nob)))
            f.seek(8, 1)

        hz = np.fromfile(f, dtype=FLOAT, count=int(nx * ny)).reshape((ny, nx))
        f.seek(8, 1)
        mz = np.fromfile(f, dtype=INT, count=int(nx * ny)).reshape((ny, nx))

    hz = hz.transpose()
    mz = mz.transpose()

    lon_z, lat_z = arakawa_grid(nx, ny, x0, x1, y0, y1, "h")
    lon_u, lat_u = arakawa_grid(nx, ny, x0, x1, y0, y1, "u")
    lon_v, lat_v = arakawa_grid(nx, ny, x0, x1, y0, y1, "v")
    lat_z, lon_z = np.meshgrid(lat_z, lon_z)
    lat_u, lon_u = np.meshgrid(lat_u, lon_u)
    lat_v, lon_v = np.meshgrid(lat_v, lon_v)

    # Dataset Otis style
    dset = xr.Dataset()

    # Coords
    dset["lon_z"] = xr.DataArray(da.from_array(lon_z), dims=("nx", "ny"))
    dset["lat_z"] = xr.DataArray(da.from_array(lat_z), dims=("nx", "ny"))
    dset["lon_u"] = xr.DataArray(da.from_array(lon_u), dims=("nx", "ny"))
    dset["lat_u"] = xr.DataArray(da.from_array(lat_u), dims=("nx", "ny"))
    dset["lon_v"] = xr.DataArray(da.from_array(lon_v), dims=("nx", "ny"))
    dset["lat_v"] = xr.DataArray(da.from_array(lat_v), dims=("nx", "ny"))

    # Mask
    dset["mz"] = xr.DataArray(da.from_array(mz), dims=("nx", "ny")).astype("int32")
    dset["mu"] = dset.mz * dset.mz.roll(nx=1, roll_coords=False)
    dset["mv"] = dset.mz * dset.mz.roll(ny=1, roll_coords=False)

    # Depth
    dset["hz"] = xr.DataArray(da.from_array(hz), dims=("nx", "ny")).astype("float64")
    dset["hu"] = u_from_z(dset.hz, dset.mz, dset.mu)
    dset["hv"] = v_from_z(dset.hz, dset.mz, dset.mv)

    # Indices open boundaries
    # dset["iob_z"] = xr.DataArray(da.from_array(iob), dims=("iiob", "nob_z"))
    dset["iob_z"] = xr.DataArray(indices_open_boundary(dset.mz), dims=("iiob", "nob_z"))
    dset["iob_u"] = xr.DataArray(indices_open_boundary(dset.mu), dims=("iiob", "nob_u"))
    dset["iob_v"] = xr.DataArray(indices_open_boundary(dset.mv), dims=("iiob", "nob_v"))

    # Attributes
    set_attributes(dset, "otis")
    dset.attrs = {
        "type": "OTIS Arakawa C-grid file",
        "title": "Oceantide bathymetry from binary file",
    }

    return dset


def write_otis_bin_u(ufile, URe, UIm, VRe, VIm, con, lon, lat):
    """Write elevation constituents data in the otis binary file.

    Args:
        - ufile (str): Name of transports binary constituents file to write.
        - URe (DataArray): Real eastward transport :math:`\\Re{U}(nc,nx,ny)`.
        - UIm (DataArray): Imag eastward transport :math:`\\Im{U}(nc,nx,ny)`.
        - VRe (DataArray): Real northward transport :math:`\\Re{V}(nc,nx,ny)`.
        - VIm (DataArray): Imag northward transport :math:`\\Im{V}(nc,nx,ny)`.
        - con (1darray): Constituents names (lowercase).
        - lon (DataArray, 1darray): Longitude coordinates at the cell centre (Z-nodes).
        - lat (DataArray, 1darray): Latitude coordinates at the cell centre (Z-nodes).

    Note:
        - Arrays must have shape consistent with Otis convention :math:`(nc,nx,ny)`.
        - Coordinates lon and lat can be 1d or 2d arrays.
        - Masked values are filled with zeros before writing.

    """
    nc, nx, ny = URe.shape
    URe = URe.fillna(0.0)
    UIm = UIm.fillna(0.0)
    VRe = VRe.fillna(0.0)
    VIm = VIm.fillna(0.0)

    with open(ufile, "wb") as fid:
        # Header
        delim = np.array(4 * (nc + 7), dtype=INT)
        delim.tofile(fid)
        np.array(nx, dtype=INT).tofile(fid)
        np.array(ny, dtype=INT).tofile(fid)
        np.array(nc, dtype=INT).tofile(fid)
        theta_lim(lon, lat).tofile(fid)
        con.values.astype("S4").tofile(fid)
        delim.tofile(fid)

        # Records
        delim = np.array(2 * 8 * nx * ny, dtype=INT)
        for ic in range(nc):
            delim.tofile(fid)
            data = np.zeros((ny, nx * 4))
            data[:, 0 : 4 * nx - 3 : 4] = URe[ic].T
            data[:, 1 : 4 * nx - 2 : 4] = UIm[ic].T
            data[:, 2 : 4 * nx - 1 : 4] = VRe[ic].T
            data[:, 3 : 4 * nx - 0 : 4] = VIm[ic].T
            data.astype(FLOAT).tofile(fid)
            delim.tofile(fid)


def write_otis_bin_h(hfile, hRe, hIm, con, lon, lat):
    """Write elevation constituents data in the otis binary file.

    Args:
        - hfile (str): Name of elevation binary constituents file to write.
        - hRe (DataArray): Real elevation :math:`\\Re{h}(nc,nx,ny)`.
        - hIm (DataArray): Imag elevation :math:`\\Im{h}(nc,nx,ny)`.
        - con (DataArray, 1darray): Constituents names.
        - lon (DataArray, 1darray): Longitude coordinates at the cell centre (Z-nodes).
        - lat (DataArray, 1darray): Latitude coordinates at the cell centre (Z-nodes).

    Note:
        - Arrays must have shape consistent with Otis convention :math:`(nc,nx,ny)`.
        - Coordinates lon and lat can be 1d or 2d arrays.
        - Masked values are filled with zeros before writing.

    """
    nc, nx, ny = hRe.shape

    hRe = hRe.fillna(0.0)
    hIm = hIm.fillna(0.0)

    # Cons array in bytes format
    if isinstance(con, xr.DataArray):
        con = con.values
    if isinstance(con[0], np.str_):
        con = [f"{c.lower():4s}" for c in con]
    elif isinstance(con[0], np.bytes_):
        con = [f"{c.decode('utf-8').lower():4s}" for c in con]

    with open(hfile, "wb") as fid:
        # Header
        delim = np.array(4 * (nc + 7), dtype=INT)
        delim.tofile(fid)
        np.array(nx, dtype=INT).tofile(fid)
        np.array(ny, dtype=INT).tofile(fid)
        np.array(nc, dtype=INT).tofile(fid)
        theta_lim(lon, lat).tofile(fid)
        np.array(con, dtype="S4").tofile(fid)
        delim.tofile(fid)

        # Records
        delim = np.array(8 * nx * ny, dtype=INT)
        for ic in range(nc):
            delim.tofile(fid)
            data = np.zeros((ny, nx * 2))
            data[:, 0 : 2 * nx - 1 : 2] = hRe[ic].T
            data[:, 1 : 2 * nx - 0 : 2] = hIm[ic].T
            data.astype(FLOAT).tofile(fid)
            delim.tofile(fid)


def write_otis_bin_grid(gfile, hz, mz, lon, lat, dt=12):
    """Write grid data in the otis binary file.

    Args:
        - gfile (str): Name of grid binary file to write.
        - hz (DataArray): Water depth at Z-nodes :math:`h(nx,ny)`.
        - mz (DataArray): Land mask at Z-nodes :math:`m(nx,ny)`.
        - lon (DataArray, 1darray): Longitude coordinates at the cell centre (Z-nodes).
        - lat (DataArray, 1darray): Latitude coordinates at the cell centre (Z-nodes).

    Note:
        - Arrays must have shape consistent with Otis convention :math:`(nc,nx,ny)`.
        - Coordinates lon and lat can be 1d or 2d arrays.

    """
    nx, ny = hz.shape
    iob = indices_open_boundary(mz)
    nob = iob.shape[1]

    with open(gfile, "wb") as fid:
        # Header
        delim = np.array(32, dtype=INT)
        delim.tofile(fid)
        np.array([nx, ny], dtype=INT).tofile(fid)
        theta_lim(lon, lat).tofile(fid)
        np.array(dt, dtype=FLOAT).tofile(fid)
        np.array(nob, dtype=INT).tofile(fid)
        delim.tofile(fid)

        # Indices open boundaries
        delim = np.array(8 * nob, dtype=INT)
        if nob == 0:
            np.array([4, 0, 4], dtype=INT).tofile(fid)
        else:
            delim.tofile(fid)
            iob.tofile(fid)
            delim.tofile(fid)

        # Water depth
        delim = np.array(4 * nx * ny, dtype=INT)
        delim.tofile(fid)
        hz.fillna(0.0).values.T.astype(FLOAT).tofile(fid)
        delim.tofile(fid)

        # Land mask
        delim.tofile(fid)
        mz.fillna(1.0).values.T.astype(INT).tofile(fid)
        delim.tofile(fid)


def otis_to_oceantide(dsg, dsh, dsu):
    """Convert otis datasets into oceantide format."""
    otis = OtisFormatter(dsg, dsh, dsu)
    ds = otis()
    return ds


class OtisFormatter:
    """Format Otis datasets into Oceantide.

    Args:
        - dsg (Dataset): Otis grid dataset.
        - dsh (Dataset): Otis elevation dataset.
        - dsu (Dataset): Otis transports dataset.

    Note:
        - Input datasets must have SI units.

    """

    def __init__(self, dsg, dsh, dsu):
        self.dsg = dsg
        self.dsh = dsh
        self.dsu = dsu

    def __repr__(self):
        return re.sub(r"<.+>", f"<{self.__class__.__name__}>", str(self.ds))

    def __call__(self):
        self.validate()
        self.construct()
        return self.ds

    def _to_complex(self):
        """Merge real and imaginary components into a complex variable."""
        for v in ["h", "u", "v"]:
            real = self.ds[f"{v}Re"].astype("float32")
            imag = self.ds[f"{v}Im"].astype("float32") * 1j
            self.ds[f"{v}"] = real + imag
            self.ds = self.ds.drop_vars([f"{v}Re", f"{v}Im"])

    def _to_single_grid(self):
        """Convert Arakawa into single grid at Z-nodes."""
        self.ds = xr.Dataset()

        # Variables at cell centre
        self.ds["dep"] = self.dsg.hz
        self.ds["hRe"] = self.dsh.hRe
        self.ds["hIm"] = self.dsh.hIm
        self.ds["uRe"] = z_from_u(self.dsu.URe / self.dsg.hu, self.dsg.mu, self.dsg.mz)
        self.ds["uIm"] = z_from_u(self.dsu.UIm / self.dsg.hu, self.dsg.mu, self.dsg.mz)
        self.ds["vRe"] = z_from_v(self.dsu.VRe / self.dsg.hv, self.dsg.mv, self.dsg.mz)
        self.ds["vIm"] = z_from_v(self.dsu.VIm / self.dsg.hv, self.dsg.mv, self.dsg.mz)

        # Coordinates
        self.ds = self.ds.rename({"nc": "con", "nx": "lon", "ny": "lat"})
        self.ds["con"] = self.dsh.con.rename({"nc": "con"})
        self.ds["lon"] = self.dsh.lon_z.isel(ny=0).rename({"nx": "lon"})
        self.ds["lat"] = self.dsh.lat_z.isel(nx=0).rename({"ny": "lat"})

        # Transpose
        self.ds = self.ds.transpose("con", "lat", "lon")

    def _format_cons(self):
        """Format constituents coordinates."""
        decoded = [c.upper() for c in self.ds.con.values.tobytes().decode().split()]
        self.ds = self.ds.assign_coords({"con": np.array(decoded).astype("U4")})

    def _set_attributes(self):
        """Define attributes for formatted dataset."""
        self.ds.dep.attrs = {}
        set_attributes(self.ds, "dataset")
        self.ds.attrs = {"description": "Oceantide tidal constituents"}

    def _mask_land(self):
        """Mask grid points where depth <= 0."""
        for v in ["h", "u", "v"]:
            self.ds[v] = self.ds[v].where(self.ds.dep > 0)

    def validate(self):
        """Check that input dataset has all requirements."""
        for v in ["hRe", "hIm"]:
            if v not in self.dsh.data_vars:
                raise ValueError(f"Variable {v} required in elevation dataset dsh.")
        for v in ["URe", "UIm", "VRe", "VIm"]:
            if v not in self.dsu.data_vars:
                raise ValueError(f"Variable {v} required in transports dataset dsu.")
        for v in ["hz", "hu", "hv", "mz", "mu", "mv"]:
            if v not in self.dsg.data_vars:
                raise ValueError(f"Variable {v} required in grid dataset dsg.")
        if self.dsh.con.dtype != np.dtype("S4"):
            raise ValueError(f"Constituents variables dtype must be 'S4'.")

    def construct(self):
        """Define constituents dataset."""
        self._to_single_grid()
        self._format_cons()
        self._to_complex()
        self._set_attributes()
        self._mask_land()



"""
This tidal ellipse package is taken from the PyFVCOM module written Created by
Pierre Cazenave available at https://github.com/pwcazenave/PyFVCOM.
The author retains the copyright of this program.
"""

"""
Python translation of Zhigang Xu's tidal_ellipse MATLAB tools, available at
http://woodshole.er.usgs.gov/operations/sea-mat/tidal_ellipse-html/ap2ep.html

Converted to Python by Pierre Cazenave (Plymouth Marine Laboratory), October
2012. Email: pica@pml.ac.uk

Authorship Copyright:

   The author retains the copyright of this program, while you are welcome
to use and distribute it as long as you credit the author properly and
respect the program name itself. Particularly, you are expected to retain the
original author's name in this original version or any of its modified version
that you might make. You are also expected not to essentially change the name
of the programs except for adding possible extension for your own version you
might create, e.g. ap2ep_xx is acceptable.  Any suggestions are welcome and
enjoy my program(s)!


Author Info:
_______________________________________________________________________________

  Zhigang Xu, Ph.D.
  (pronounced as Tsi Gahng Hsu)
  Research Scientist
  Coastal Circulation
  Bedford Institute of Oceanography
  1 Challenge Dr.
  P.O. Box 1006                    Phone  (902) 426-2307 (o)
  Dartmouth, Nova Scotia           Fax    (902) 426-7827
  CANADA B2Y 4A2                   email xuz@dfo-mpo.gc.ca
_______________________________________________________________________________

Release Date: Nov. 2000, Revised on May. 2002 to adopt Foreman's northern semi
major axis convention.

"""

import numpy as np
import matplotlib.pyplot as plt


def ap2ep(Au, PHIu, Av, PHIv, plot_demo=False):
    """
    Convert tidal amplitude and phase lag (ap-) parameters into tidal ellipse
    (ep-) parameters. Please refer to ep2ap for its inverse function.

    Usage:

    SEMA, ECC, INC, PHA, w = ap2ep(Au, PHIu, Av, PHIv, plot_demo=False)

    Where:

        Au, PHIu, Av, PHIv are the amplitudes and phase lags (in degrees) of
        u- and v- tidal current components. They can be vectors or
        matrices or multidimensional arrays.

        plot_demo is an optional argument, when it is supplied as an array
        of indices, say [i j k l], the program will plot an ellipse
        corresponding to Au[i, j, k, l], PHIu[i, j, k, l], Av[i, j, k, l], and
        PHIv[i, j, k, l]. Defaults to False (i.e. no plot).

        Any number of dimensions are allowed as long as your computer
        resource can handle.

        SEMA: Semi-major axes, or the maximum speed.

        ECC:  Eccentricity, the ratio of semi-minor axis over the semi-major
              axis; its negative value indicates that the ellipse is traversed
              in clockwise direction.

        INC:  Inclination, the angles (in degrees) between the semi-major axes
              and u-axis.

        PHA:  Phase angles, the time (in angles and in degrees) when the tidal
              currents reach their maximum speeds,  (i.e.  PHA=omega*tmax).

              These four ep-parameters will have the same dimensionality (i.e.,
              vectors, or matrices) as the input ap-parameters.

        w:    A matrix whose rows allow for plotting ellipses and whose columns
              are for different ellipses corresponding columnwise to SEMA. For
              example, plot(np.real(w[0, :]), np.imag(w[0, :])) will let you
              see the first ellipse. You may need to use squeeze function when
              w is a more than two dimensional array. See example.py.

    Document:   tidal_ellipse.ps

    Revisions: May  2002, by Zhigang Xu,  --- adopting Foreman's northern semi
    major axis convention.

    For a given ellipse, its semi-major axis is undetermined by 180. If we
    borrow Foreman's terminology to call a semi major axis whose direction lies
    in a range of [0, 180) as the northern semi-major axis and otherwise as a
    southern semi major axis, one has freedom to pick up either northern or
    southern one as the semi major axis without affecting anything else.
    Foreman (1977) resolves the ambiguity by always taking the northern one as
    the semi-major axis. This revision is made to adopt Foreman's convention.
    Note the definition of the phase, PHA, is still defined as the angle
    between the initial current vector, but when converted into the maximum
    current time, it may not give the time when the maximum current first
    happens; it may give the second time that the current reaches the maximum
    (obviously, the 1st and 2nd maximum current times are half tidal period
    apart) depending on where the initial current vector happen to be and its
    rotating sense.

    Version 2, May 2002

    Converted to Python by Pierre Cazenave, October 2012.

    Authorship Copyright:

       The author retains the copyright of this program, while  you are welcome
    to use and distribute it as long as you credit the author properly and
    respect the program name itself. Particularly, you are expected to retain
    the original author's name in this original version or any of its modified
    version that you might make. You are also expected not to essentially
    change the name of the programs except for adding possible extension for
    your own version you might create, e.g. ap2ep_xx is acceptable.  Any
    suggestions are welcome and enjoy my program(s)!


    Author Info:
    _______________________________________________________________________
      Zhigang Xu, Ph.D.
      (pronounced as Tsi Gahng Hsu)
      Research Scientist
      Coastal Circulation
      Bedford Institute of Oceanography
      1 Challenge Dr.
      P.O. Box 1006                    Phone  (902) 426-2307 (o)
      Dartmouth, Nova Scotia           Fax    (902) 426-7827
      CANADA B2Y 4A2                   email xuz@dfo-mpo.gc.ca
    _______________________________________________________________________

    Release Date: Nov. 2000, Revised on May. 2002 to adopt Foreman's northern
    semi major axis convention.

    """

    # Assume the input phase lags are in degrees and convert them in radians.
    PHIu = PHIu / 180 * np.pi
    PHIv = PHIv / 180 * np.pi

    # Make complex amplitudes for u and v
    i = 1j
    u = Au * np.exp(-i * PHIu)
    v = Av * np.exp(-i * PHIv)

    # Calculate complex radius of anticlockwise and clockwise circles:
    wp = (u + i * v) / 2  # for anticlockwise circles
    wm = np.conj(u - i * v) / 2  # for clockwise circles
    # and their amplitudes and angles
    Wp = np.abs(wp)
    Wm = np.abs(wm)
    THETAp = np.angle(wp)
    THETAm = np.angle(wm)

    # calculate ep-parameters (ellipse parameters)
    SEMA = Wp + Wm  # Semi Major Axis, or maximum speed
    SEMI = Wp - Wm  # Semi Minor Axis, or minimum speed
    ECC = SEMI / SEMA  # Eccentricity

    PHA = (THETAm - THETAp) / 2  # Phase angle, the time (in angle) when
    # the velocity reaches the maximum
    INC = (THETAm + THETAp) / 2  # Inclination, the angle between the
    # semi major axis and x-axis (or u-axis).

    # convert to degrees for output
    PHA = PHA / np.pi * 180
    INC = INC / np.pi * 180
    THETAp = THETAp / np.pi * 180
    THETAm = THETAm / np.pi * 180

    # map the resultant angles to the range of [0, 360].
    PHA = np.mod(PHA + 360, 360)
    INC = np.mod(INC + 360, 360)

    # Mar. 2, 2002 Revision by Zhigang Xu    (REVISION_1)
    # Change the southern major axes to northern major axes to conform the tidal
    # analysis convention  (cf. Foreman, 1977, p. 13, Manual For Tidal Currents
    # Analysis Prediction, available in www.ios.bc.ca/ios/osap/people/foreman.htm)
    k = np.fix(INC / 180)
    INC = INC - k * 180
    PHA = PHA + k * 180
    PHA = np.mod(PHA, 360)

    if plot_demo:
        plot_ell(SEMA, ECC, INC, PHA)

    ndot = np.prod(np.shape(SEMA))
    dot = 2 * np.pi / ndot
    ot = np.arange(0, 2 * np.pi, dot)
    w = wp.flatten() * np.exp(i * ot) + wm.flatten() * np.exp(-i * ot)
    w = np.reshape(w, np.shape(wp))

    return SEMA, ECC, INC, PHA, w


def ep2ap(SEMA, ECC, INC, PHA, plot_demo=False):
    """
    Convert tidal ellipse parameters into amplitude and phase lag parameters.
    Its inverse is app2ep.m. Please refer to app2ep for the meaning of the
    inputs and outputs.

    Zhigang Xu
    Oct. 20, 2000

    Converted to Python by Pierre Cazenave, October 2012.

    Authorship Copyright:

        The author of this program retains the copyright of this program, while
    you are welcome to use and distribute this program as long as you credit
    the author properly and respect the program name itself. Particularly, you
    are expected to retain the original author's name in this original version
    of the program or any of its modified version that you might make.  You are
    also expected not to essentially change the name of the programs except for
    adding possible extension for your own version you might create, e.g.
    app2ep_xx is acceptable.  Any suggestions are welcome and enjoy my
    program(s)!


    Author Info:
    _______________________________________________________________________
      Zhigang Xu, Ph.D.
      (pronounced as Tsi Gahng Hsu)
      Research Scientist
      Coastal Circulation
      Bedford Institute of Oceanography
      1 Challenge Dr.
      P.O. Box 1006                    Phone  (902) 426-2307 (o)
      Dartmouth, Nova Scotia           Fax    (902) 426-7827
      CANADA B2Y 4A2                   email xuz@dfo-mpo.gc.ca
    _______________________________________________________________________

    Release Date: Nov. 2000

    """

    i = 1j

    Wp = (1 + ECC) / 2 * SEMA
    Wm = (1 - ECC) / 2 * SEMA
    THETAp = INC - PHA
    THETAm = INC + PHA

    # Convert degrees into radians
    THETAp = THETAp / 180 * np.pi
    THETAm = THETAm / 180 * np.pi

    # Calculate wp and wm.
    wp = Wp * np.exp(i * THETAp)
    wm = Wm * np.exp(i * THETAm)

    ndot = np.prod(np.shape(SEMA))
    dot = 2 * np.pi / ndot
    ot = np.arange(0, 2 * np.pi, dot)
    w = wp.flatten() * np.exp(i * ot) + wm.flatten() * np.exp(-i * ot)
    w = np.reshape(w, np.shape(wp))

    # Calculate cAu, cAv --- complex amplitude of u and v
    cAu = wp + np.conj(wm)
    cAv = -i * (wp - np.conj(wm))
    Au = np.abs(cAu)
    Av = np.abs(cAv)
    PHIu = -np.angle(cAu) * 180 / np.pi
    PHIv = -np.angle(cAv) * 180 / np.pi

    # flip angles in the range of [-180 0) to the range of [180 360).
    id = PHIu < 0
    PHIu[id] = PHIu[id] + 360
    id = PHIv < 0
    PHIv[id] = PHIv[id] + 360

    if plot_demo:
        plot_ell(SEMA, ECC, INC, PHA, plot_demo)

    return Au, PHIu, Av, PHIv, w


def cBEpm(g, f, sigma, nu, kappa, z, h):
    """
    Evaluate the theoretical vertical profiles (or Bottom Ekman spiral
    profiles) of tidal currents in the two rotary directions driven by
    half-unit of sea surface gradients in the two directions respectively. Eddy
    viscosity is assumed as vertically invariant. See tidal_ellipse.ps for more
    details.

    Inputs:

        g:      acceleration gravity
        f:      the Coriolis parameter
        nu:     the eddy viscosity
        kappa:  the bottom frictional coefficient
        z:      the vertical coordinates, can be a vector but must be
                within [0 -h];
        h:      the water depth, must be positive.

        Note: except for z, all other inputs must be scalars.

    Outputs:

        BEp and BEm, the same dimensions of z,  the outputs for the vertical
            velocity profiles driven respectively by a unit of sea surface
            slope in the positive rotation direction and negative rotation
            direction for when the eddy viscosity is vertically invariant. See
            the associated document for more details.

    Authorship Copyright:

       The author of this program retains the copyright of this program, while
    you are welcome to use and distribute this program as long as you credit
    the author properly and respect the program name itself. Particularly,
    you are expected to retain the original author's name in this original
    version of the program or any of its modified version that you might make.
    You are also expected not to essentially change the name of the programs
    except for adding possible extension for your own version you might create,
    e.g. ap2ep_xx is acceptable.  Any suggestions are welcome and enjoying my
    program(s)!


    Author Info:
    _______________________________________________________________________
      Zhigang Xu, Ph.D.
      (pronounced as Tsi Gahng Hsu)
      Research Scientist
      Coastal Circulation
      Bedford Institute of Oceanography
      1 Challenge Dr.
      P.O. Box 1006                    Phone  (902) 426-2307 (o)
      Dartmouth, Nova Scotia           Fax    (902) 426-7827
      CANADA B2Y 4A2                   email zhigangx@emerald.bio.dfo.ca
                                             zhigang_xu_98@yahoo.com
    _______________________________________________________________________

    Release Date: Nov. 2000

    """

    if (
        (len(g) > 1)
        | (len(f) > 1)
        | (len(sigma) > 1)
        | (len(nu) > 1)
        | (len(kappa) > 1)
        | (len(h) > 1)
    ):
        print("inputs of g, f, sigma, nu, kappa, and h should be all scalars!")
        raise

    if (any(z / h > 0)) | (any(z / h < -1)):
        print("z must be negative and must be within [0 -h]")

    delta_e = np.sqrt(2 * nu / f)  # Ekman depth
    alpha = (1 + 1j) / delta_e * np.sqrt(1 + sigma / f)
    beta = (1 + 1j) / delta_e * np.sqrt(1 - sigma / f)

    BEp = get_BE(g, alpha, h, z, nu, kappa)
    BEm = get_BE(g, beta, h, z, nu, kappa)

    return BEp, BEm


def get_BE(g, alpha, h, z, nu, kappa):
    """Child function of cBEpm"""

    z = z.flatten()
    z_h = z / h
    ah = alpha * h
    az = alpha * z
    ah2 = ah * 2
    anu_k = alpha * nu / kappa
    nu_kh = nu / (kappa * h)

    # Series solution
    if abs(ah) < 1:
        T = 10
        C = -g * h * h / (nu * (1 + anu_k * np.tanh(ah))) * 2
        A1 = (1 - z_h * z_h) / 2 + nu_kh
        B1 = np.exp(-ah) / (1 + np.exp(-ah2))
        B = B1
        series_sum = A1 * B1

        for t in np.arange(2, T):
            t2 = 2 * t
            A = (1 - z_h**t2) / t2 + nu_kh
            B = B * ah * ah / (t2 - 1) / (t2 - 2)
            series_sum = series_sum + A * B

        BE = C * series_sum

    # Finite solution
    else:
        c = -g * h * h / nu
        denom = (np.exp(az - ah) + np.exp(-(az + ah))) / (1 + np.exp(-2 * ah))

        numer = 1 + anu_k * np.tanh(ah)
        BE = c * ((1 - denom / numer) / (ah * ah))

    return BE


def sub2ind(shape, pos):
    """
    Substitute of MATLAB's sub2ind function for NumPy.

    t = numpy.random.random([2, 4, 5, 2])
    n = sub2ind(numpy.shape(t), [1, 2, 4, 1])
    >>> n
    69

    From http://stackoverflow.com/questions/4114461

    """
    res = 0
    acc = 1
    for pi, si in zip(reversed(pos), reversed(shape)):
        res += pi * acc
        acc *= si

    return res


def plot_ell(SEMA, ECC, INC, PHA, IND=[1]):
    """
    An auxiliary function used in ap2ep and ep2ap for plotting tidal ellipse.
    The inputs, MA, ECC, INC and PHA are the output of ap2ep and IND is a
    vector for indices for plotting a particular ellipse, e.g., if IND=[2 3 1]
    the ellipse corresponding to the indices of [2,3,1] will be plotted.

    By default, the first ellipse is always plotted.

    Converted to Python by Pierre Cazenave, October 2012.

    """

    len_IND = len(IND)
    if IND:
        cmd = "sub2ind(SEMA.shape, "
        if len_IND == 1:
            titletxt = "Ellipse "
        else:
            titletxt = "Ellipse ("

        for k in range(len_IND):
            if k == 0:
                cmd = cmd + "[" + str(IND[k])
            else:
                cmd = cmd + "," + str(IND[k])

            if k < len_IND - 1:
                titletxt = titletxt + str(IND[k]) + ","
            elif len_IND == 1:
                titletxt = titletxt + str(IND[k])
            else:
                titletxt = titletxt + str(IND[k]) + ")"

        cmd = "n = " + cmd + "])"
        # This is pretty nasty, but it works.
        exec(cmd)

        plt.gcf()
        plt.clf()
        do_the_plot(
            SEMA.flatten()[n], ECC.flatten()[n], INC.flatten()[n], PHA.flatten()[n]
        )
        titletxt = titletxt + ",  (red) green (anti-) clockwise component"
        plt.title(titletxt)
    elif len_IND:
        print("IND input contains zero element(s)!\nNo ellipse will be plotted.")


def do_the_plot(SEMA, ECC, INC, PHA):
    """
    Ellipse plot subfunction.

    Converted to Python by Pierre Cazenave, October 2012.

    """

    i = 1j

    SEMI = SEMA * ECC
    Wp = (1 + ECC) / 2 * SEMA
    Wm = (1 - ECC) / 2 * SEMA
    THETAp = INC - PHA
    THETAm = INC + PHA

    # Convert degrees into radians
    THETAp = THETAp / 180 * np.pi
    THETAm = THETAm / 180 * np.pi
    INC = INC / 180 * np.pi
    PHA = PHA / 180 * np.pi

    # Calculate wp and wm.
    wp = Wp * np.exp(i * THETAp)
    wm = Wm * np.exp(i * THETAm)

    dot = np.pi / 36
    ot = np.arange(0, 2 * np.pi, dot)
    a = wp * np.exp(i * ot)
    b = wm * np.exp(-i * ot)
    w = a + b

    wmax = SEMA * np.exp(i * INC)
    wmin = SEMI * np.exp(i * (INC + np.pi / 2))

    plt.plot(np.real(w), np.imag(w))
    plt.axis("equal")
    plt.hold("on")
    plt.plot([0, np.real(wmax)], [0, np.imag(wmax)], "m")
    plt.plot([0, np.real(wmin)], [0, np.imag(wmin)], "m")
    plt.xlabel("u")
    plt.ylabel("v")
    plt.plot(np.real(a), np.imag(a), "r")
    plt.plot(np.real(b), np.imag(b), "g")
    plt.plot([0, np.real(a[0])], [0, np.imag(a[0])], "ro")
    plt.plot([0, np.real(b[0])], [0, np.imag(b[0])], "go")
    plt.plot([0, np.real(w[0])], [0, np.imag(w[0])], "bo")
    plt.plot(np.real(a[0]), np.imag(a[0]), "ro")
    plt.plot(np.real(b[0]), np.imag(b[0]), "go")
    plt.plot(np.real(w[0]), np.imag(w[0]), "bo")
    plt.plot(
        np.real([a[0], a[0] + b[0]]),
        np.imag([a[0], a[0] + b[0]]),
        linestyle="--",
        color="g",
    )
    plt.plot(
        np.real([b[0], a[0] + b[0]]),
        np.imag([b[0], a[0] + b[0]]),
        linestyle="--",
        color="r",
    )

    for n in range(len(ot)):
        plt.hold("on")
        plt.plot(np.real(a[n]), np.imag(a[n]), "ro")
        plt.plot(np.real(b[n]), np.imag(b[n]), "go")
        plt.plot(np.real(w[n]), np.imag(w[n]), "bo")

    plt.hold("off")
    plt.show()


def prep_plot(SEMA, ECC, INC, PHA):
    """
    Take the output of ap2ep (SEMA, ECC, INC, and PHA) and prepare it for
    plotting.

    This is extracted from do_the_plot above, but allows quicker access when
    all that is required is a plot of an ellipse, for which only w is really
    required.

    Returns w, wmin and wmax (w is used for plotting the ellipse, see
    plot_ell).

    """

    i = 1j

    SEMI = SEMA * ECC
    Wp = (1 + ECC) / 2 * SEMA
    Wm = (1 - ECC) / 2 * SEMA
    THETAp = INC - PHA
    THETAm = INC + PHA

    # Convert degrees into radians
    THETAp = THETAp / 180 * np.pi
    THETAm = THETAm / 180 * np.pi
    INC = INC / 180 * np.pi
    PHA = PHA / 180 * np.pi

    # Calculate wp and wm.
    wp = Wp * np.exp(i * THETAp)
    wm = Wm * np.exp(i * THETAm)

    dot = np.pi / 36
    ot = np.arange(0, 2 * np.pi, dot)
    a = wp * np.exp(i * ot)
    b = wm * np.exp(-i * ot)
    w = a + b

    # Repeat the first position in w so we close the ellipse.
    w = np.hstack((w, w[0]))

    wmax = SEMA * np.exp(i * INC)
    wmin = SEMI * np.exp(i * (INC + np.pi / 2))

    return w, wmin, wmax


if __name__ == "__main__":
    """
    Replicate the tidal ellipse example file from Zhigang Xu's tidal_ellipse
    MATLAB toolbox.

    Pierre Cazenave (Plymouth Marine Laboratory), October 2012.

    """

    # Demonstrate how to use ap2ep and ep2ap
    Au = np.random.random([4, 3, 2])  # so 4x3x2 multi-dimensional matrices
    Av = np.random.random([4, 3, 2])  # are used for the demonstration.
    Phi_v = np.random.random([4, 3, 2]) * 360  # phase lags inputs are expected to
    Phi_u = np.random.random([4, 3, 2]) * 360  # be in degrees.

    plt.figure(1)
    plt.clf()
    SEMA, ECC, INC, PHA, w = ap2ep(Au, Phi_u, Av, Phi_v, [2, 3, 1])
    plt.figure(2)
    plt.clf()
    rAu, rPhi_u, rAv, rPhi_v, rw = ep2ap(SEMA, ECC, INC, PHA, [2, 3, 1])

    # Check if ep2ap has recovered Au, Phi_u, Av, Phi_v
    print((np.max(np.abs(rAu - Au).flatten())))  # = 9.9920e-16, = 2.22044604925e-16
    print((np.max(np.abs(rAv - Av).flatten())))  # = 6.6613e-16, = 7.77156117238e-16
    print(
        (np.max(np.abs(rPhi_u - Phi_u).flatten()))
    )  # = 4.4764e-13, = 1.70530256582e-13
    print(
        (np.max(np.abs(rPhi_v - Phi_v).flatten()))
    )  # = 1.1369e-13, = 2.27373675443e-13
    print(
        (np.max(np.max(np.abs(w - rw).flatten())))
    )  # = 1.3710e-15, = 1.1322097734e-15
    # For the random realization I (Zhigang Xu) had, the differences are listed
    # on the right hand of the above column. I (Pierre Cazenave) got the second
    # column with the Python version. What are yours?

    # Zhigang Xu
    # Nov. 12, 2000
    # Pierre Cazenave
    # October, 2012



"""Otis binary output."""
from pathlib import Path
import numpy as np

from oceantide.core.otis import (
    write_otis_bin_h,
    write_otis_bin_u,
    write_otis_bin_grid,
    u_from_z,
    v_from_z,
)


def to_otis_binary(
    self,
    dirname: str,
    hfile: bool = True,
    ufile: bool = True,
    gfile: bool = True,
    suffix: str = None,
) -> dict:
    """Write dataset as Otis binary format.

    Parameters
    ----------
    self (oceantide.tide.Tide)
        Oceantide Tide instance.
    dirname (str)
        Directory to save binary files.
    hfile (bool)
        Save tidal elevation binary file.
    ufile (bool)
        Save tidal transports binary file.
    gfile (bool)
        Save grid file.
    suffix (str)
        Suffix to define file names, by default defined by cons names.

    Returns
    -------
    filename (dict)
        Name of files written.

    """
    ds = self._obj.transpose("con", "lon", "lat", ...)
    ds = ds.rename({"con": "nc", "lon": "nx", "lat": "ny"})
    ds = ds.fillna(0.0)

    if suffix is None:
        suffix = "".join(list(ds.nc.values)).lower()

    filenames = {}

    # Write elevations
    if hfile:
        fname = f"h_{suffix.lstrip('_')}" if suffix else "h"
        filenames["hfile"] = Path(dirname) / fname
        write_otis_bin_h(
            hfile=filenames["hfile"],
            hRe=ds.h.real,
            hIm=ds.h.imag,
            con=ds.nc,
            lon=ds.nx,
            lat=ds.ny,
        )

    # Write transports
    if ufile:
        fname = f"u_{suffix.lstrip('_')}" if suffix else "u"
        filenames["ufile"] = Path(dirname) / fname
        mz = ds.dep > 0
        mu = mz * mz.roll(nx=1, roll_coords=False)
        mv = mz * mz.roll(ny=1, roll_coords=False)
        write_otis_bin_u(
            ufile=filenames["ufile"],
            URe=u_from_z(ds.u.real * ds.dep, mz, mu),
            UIm=u_from_z(ds.u.imag * ds.dep, mz, mu),
            VRe=v_from_z(ds.v.real * ds.dep, mz, mv),
            VIm=v_from_z(ds.v.imag * ds.dep, mz, mv),
            con=ds.nc,
            lon=ds.nx,
            lat=ds.ny,
        )

    # Write grid
    if gfile:
        fname = f"grid_{suffix.lstrip('_')}" if suffix else "grid"
        filenames["gfile"] = Path(dirname) / fname
        write_otis_bin_grid(
            gfile=filenames["gfile"],
            hz=ds.dep,
            mz=mz,
            lon=ds.nx,
            lat=ds.ny,
            dt=12.0,
        )

    fname = f"model_{suffix.lstrip('_')}" if suffix else "model"
    with open(Path(dirname) / fname, mode="w") as stream:
        for filename in filenames.values():
            stream.write(f"./{filename.name}\n")

    return filenames



"""Read Otis netcdf file format."""
import os
import xarray as xr

from oceantide.core.otis import otis_to_oceantide, otis_filenames
from oceantide.tide import Tide


def read_otis_netcdf(
    filename: str = None, gfile: str = None, hfile: str = None, ufile: str = None
):
    """Read Otis Netcdf file format.

    Parameters
    ----------
    filename (str)
        Name of `Model_*` metadata file specifying other files to read.
    gfile (str)
        Name of grid file to read, by default defined from `filename`.
    hfile (str)
        Name of elevation file to read, by default defined from `filename`.
    ufile (str)
        Name of currents file to read, by default defined from `filename`.

    Returns
    -------
    dset (xr.Dataset)
        Formatted dataset with the tide accessor.

    Notes
    -----
    Otis data are usually provided in 4 separate files:

    - Meta file named as `Model_*` specifyig the 3 data files to read.
    - Grid file named as `grid*.nc` with grid information and model depths.
    - Elevation file named as `h*.*.nc` with elevation constituents data.
    - Transport file named as `uv.*.nc` with transport constituents data.

    The path of the three data files can be prescribed either by specifying the meta
    file path (`filename`) or by explicitly providing their path (`gfile`, `hfile` and
    `ufile`).

    """
    if filename is not None:
        _gfile, _hfile, _ufile = otis_filenames(filename)
    else:
        _gfile = _hfile = _ufile = None
        if not all([gfile, hfile, ufile]):
            raise ValueError(
                "Either specify `filename` or all of `gfile`, `hfile`, `ufile`."
            )

    gfile = gfile or _gfile
    hfile = hfile or _hfile
    ufile = ufile or _ufile

    dsg = xr.open_dataset(gfile, chunks={})
    dsh = xr.open_dataset(hfile, chunks={})
    dsu = xr.open_dataset(ufile, chunks={})

    dset = otis_to_oceantide(dsg, dsh, dsu)

    return dset



"""Oceantide output."""
from pathlib import Path
import numpy as np
import xarray as xr
from zarr.codecs import FixedScaleOffset

from oceantide.core.utils import set_attributes, compute_scale_and_offset


AMPMIN = -20.0
AMPMAX = 20.0
DEPMIN = 0.0
DEPMAX = 12000.0
SCALE_FACTOR_D, ADD_OFFSET_D = compute_scale_and_offset(DEPMIN, DEPMAX)
SCALE_FACTOR_A, ADD_OFFSET_A = compute_scale_and_offset(AMPMIN, AMPMAX)

FILE_FORMATS = {
    ".nc": "netcdf",
    ".zarr": "zarr",
}


def to_oceantide(self, filename: str, file_format: str = None, **kwargs):
    """Write dataset as Oceantide format.

    The oceantide format has complex constituents variables split into real and imag
    variables allowing for better packing and making it easier to support netcdf.

    Parameters
    ----------
    self (oceantide.tide.Tide)
        Oceantide Tide instance.
    filename (str)
        Name for ouput file.
    file_format (str)
        Format for output file, `zarr` and `netcdf` are supported. If not specified it
        is guessed from the filename extension.
    kwargs
        Keyword argument to pass to to_netcdf or to_zarr method.

    """
    dset = self._obj[["dep"]]
    for v in ["h", "u", "v"]:
        if v in self._obj:
            dset[f"{v}_real"] = self._obj[v].real
            dset[f"{v}_imag"] = self._obj[v].imag
    set_attributes(dset, "oceantide")

    ext = Path(filename).suffix
    if not file_format:
        try:
            file_format = FILE_FORMATS[ext]
        except KeyError as err:
            raise ValueError(
                "The filename extension must be one of ['.nc', '.zarr'] if "
                f"file_format is not specified, got '{ext}'"
            ) from err
    try:
        writer = globals()[f"_write_{file_format}"]
    except KeyError as err:
        raise ValueError(
            f"Supported file formats are ['netcdf', 'zarr'], got '{file_format}'"
        ) from err
    writer(dset, filename, **kwargs)


def _write_zarr(dset: xr.Dataset, filename: str, **kwargs):
    """Write oceantide zarr file format.

    Parameters
    ----------
    dset (xr.Dataset)
        Dataset to write.
    filename (str)
        Name for ouput file.
    kwargs
        Keyword argument to pass to to_zarr method.

    """
    kw = {"dtype": "float32", "astype": "int16"}
    fd = FixedScaleOffset(offset=ADD_OFFSET_D, scale=1 / SCALE_FACTOR_D, **kw)
    fa = FixedScaleOffset(offset=ADD_OFFSET_A, scale=1 / SCALE_FACTOR_A, **kw)

    dset.dep.encoding = {"filters": [fd], "_FillValue": DEPMAX, "dtype": kw["dtype"]}
    for varname, dvar in dset.data_vars.items():
        if varname == "dep":
            continue
        dvar.encoding = {"filters": [fa], "_FillValue": AMPMAX, "dtype": kw["dtype"]}

    dset.to_zarr(filename, **kwargs)


def _write_netcdf(dset: xr.Dataset, filename: str, **kwargs):
    """Write oceantide netcdf file format.

    Parameters
    ----------
    dset (xr.Dataset)
        Dataset to write.
    filename (str)
        Name for ouput file.
    kwargs
        Keyword argument to pass to to_netcdf method.

    """
    encoding = {"zlib": True, "_FillValue": np.int16(2**16 / 2 - 1), "dtype": "int16"}
    encd = {**encoding, **{"scale_factor": SCALE_FACTOR_D, "add_offset": ADD_OFFSET_D}}
    enca = {**encoding, **{"scale_factor": SCALE_FACTOR_A, "add_offset": ADD_OFFSET_A}}

    for varname, dvar in dset.data_vars.items():
        if varname == "dep":
            dvar.encoding = encd
        else:
            dvar.encoding = enca

    dset.to_netcdf(filename, **kwargs)



"""Read Otis Atlas netcdf file format."""
import os
import glob
from pathlib import Path
import dask
import numpy as np
import dask.array as da
import xarray as xr

from oceantide.core.otis import otis_to_oceantide
from oceantide.tide import Tide


dask.config.set({"array.slicing.split_large_chunks": False})

CONS = [
    "M2",
    "S2",
    "N2",
    "K2",
    "K1",
    "O1",
    "P1",
    "Q1",
    "MM",
    "MF",
    "M4",
    "MN4",
    "MS4",
    "2N2",
    "S1",
]


def read_otis_atlas_netcdf(
    dirname: str,
    x0: float = None,
    x1: float = None,
    y0: float = None,
    y1: float = None,
    nxchunk: int = 500,
    nychunk: int = 500,
) -> xr.Dataset:
    """Read Otis Netcdf file format.

    Parameters
    ----------
    dirname (str)
        Path name with all netcdf Atlas files. Atlas files are organised as one grid
        file and one file per constituent for uv and h.
    x0 (float)
        Longitude left corner to read.
    x1 (float)
        Longitude right corner to read.
    y0 (float)
        Latitude bottom corner to read.
    y1 (float)
        Latitude top corner to read.
    nxchunk (int)
        Chunk size along the x-axis.
    nychunk (int)
        Chunk size along the y-axis.

    Returns
    -------
    dset (xr.Dataset)
        Formatted dataset with the tide accessor.

    Notes
    -----

    - The atlas dataset is very large and requires large amount of RAM to be
      processed, we recommend setting the bounds x0, x1, y0, y1 when reading.
    - It is a good idea setting nxchunk, nychunk close to the slicing sizes.

    """
    gfile = list(Path(dirname).glob("g*.nc"))[0]
    hfile = Path(dirname).glob("h*.nc")
    ufile = Path(dirname).glob("u*.nc")

    if not hfile:
        raise ValueError(f"Cannot find elevation constituents files from {dirname}")
    if not ufile:
        raise ValueError(f"Cannot find transport constituents files from {dirname}")

    bounds = {"x0": x0, "x1": x1, "y0": y0, "y1": y1}
    dsg = read_grid(gfile, chunks={"nx": nxchunk, "ny": nychunk}, **bounds)
    dsh = read_elevations(
        hfile, chunks={"nc": None, "nx": nxchunk, "ny": nychunk}, **bounds
    )
    dsu = read_transports(
        ufile, chunks={"nc": None, "nx": nxchunk, "ny": nychunk}, **bounds
    )

    # Make coordinates 2d so they are consistent with non-atlas datasets
    dsh["lon_z"] = dsg.lon_z
    dsh["lat_z"] = dsg.lat_z
    dsu["lon_u"] = dsg.lon_u
    dsu["lat_u"] = dsg.lat_u
    dsu["lon_v"] = dsg.lon_v
    dsu["lat_v"] = dsg.lat_v

    dset = otis_to_oceantide(dsg, dsh, dsu)
    cons = [c for c in CONS if c in dset.con]
    return dset.sel(con=cons, lon=slice(x0, x1), lat=slice(y0, y1))


def indices(
    lon: np.ndarray, lat: np.ndarray, x0: float, x1: float, y0: float, y1: float
):
    """Indices of coordinates in lon and lat arrays.

    Parameters
    ----------
    lon (np.ndarray)
        Longitudes.
    lat (np.ndarray)
        Latitudes.
    x0 (float)
        Left longitude corner to read.
    x1 (float)
        Right longitude corner to read.
    y0 (float)
        Bottom latitude corner to read.
    y1 (float)
        Top latitude corner to read.

    Returns
    -------
    ix0 (int)
        Left longitude index.
    ix1 (int)
        Right longitude index.
    iy0 (int)
        Bottom latitude index.
    iy1 (int)
        Top latitude index.

    """
    ix0 = 0
    ix1 = lon.size
    iy0 = 0
    iy1 = lat.size
    if x0 is not None:
        ix0 = np.maximum(ix0, int(np.abs(lon - x0).argmin()) - 2)
    if x1 is not None:
        ix1 = np.minimum(ix1, int(np.abs(lon - x1).argmin()) + 2)
    if y0 is not None:
        iy0 = np.maximum(iy0, int(np.abs(lat - y0).argmin()) - 2)
    if y1 is not None:
        iy1 = np.minimum(iy1, int(np.abs(lat - y1).argmin()) + 2)
    return ix0, ix1, iy0, iy1


def read_grid(
    filename: str, chunks: dict, x0: float, x1: float, y0: float, y1: float
) -> xr.Dataset:
    """Read grid data.

    Parameters
    ----------
    filename (str)
        Name of grid file to read.
    chunks (dict)
        Mapping dimension names and chunking sizes.
    x0 (float)
        Left longitude corner to read.
    x1 (float)
        Right longitude corner to read.
    y0 (float)
        Bottom latitude corner to read.
    y1 (float)
        Top latitude corner to read.

    Returns
    -------
    dset (xr.Dataset)
        Merged grid dataset.

    """
    dset = xr.open_dataset(filename, chunks=chunks)

    # Slicing
    ix0, ix1, iy0, iy1 = indices(dset.lon_z, dset.lat_z, x0, x1, y0, y1)
    dset = dset.isel(nx=slice(ix0, ix1), ny=slice(iy0, iy1)).chunk(chunks)

    # Define masks
    dset["mz"] = dset.hz > 0
    dset["mu"] = dset.hu > 0
    dset["mv"] = dset.hv > 0

    # Set 2d coordinates
    lat_z, lon_z = da.meshgrid(dset.lat_z, dset.lon_z)
    lat_u, lon_u = da.meshgrid(dset.lat_u, dset.lon_u)
    lat_v, lon_v = da.meshgrid(dset.lat_v, dset.lon_v)
    dset["lon_z"] = xr.DataArray(lon_z, dims=("nx", "ny"))
    dset["lat_z"] = xr.DataArray(lat_z, dims=("nx", "ny"))
    dset["lon_u"] = xr.DataArray(lon_u, dims=("nx", "ny"))
    dset["lat_u"] = xr.DataArray(lat_u, dims=("nx", "ny"))
    dset["lon_v"] = xr.DataArray(lon_v, dims=("nx", "ny"))
    dset["lat_v"] = xr.DataArray(lat_v, dims=("nx", "ny"))
    return dset


def read_elevations(
    filenames: str, chunks: dict, x0: float, x1: float, y0: float, y1: float
) -> xr.Dataset:
    """Read and concatenate individual elevations constituents netcdf files.

    Parameters
    ----------
    filenames (list)
        Name of elevation files to read.
    chunks (dict)
        Mapping dimension names and chunking sizes.
    x0 (float)
        Left longitude corner to read.
    x1 (float)
        Right longitude corner to read.
    y0 (float)
        Bottom latitude corner to read.
    y1 (float)
        Top latitude corner to read.

    Returns
    -------
    dset (xr.Dataset)
        Merged elevation dataset.

    """
    dset = xr.open_mfdataset(
        filenames, concat_dim="nc", combine="nested", chunks=chunks
    )
    ix0, ix1, iy0, iy1 = indices(dset.lon_z, dset.lat_z, x0, x1, y0, y1)
    dset = dset.isel(nx=slice(ix0, ix1), ny=slice(iy0, iy1)).chunk(chunks)
    dset["hRe"] = dset.hRe * 1e-3
    dset["hIm"] = dset.hIm * 1e-3
    return dset


def read_transports(
    filenames: list, chunks: dict, x0: float, x1: float, y0: float, y1: float
) -> xr.Dataset:
    """Read and concatenate individual transports constituents netcdf files.

    Parameters
    ----------
    filenames (list)
        Name of transport files to read.
    chunks (dict)
        Mapping dimension names and chunking sizes.
    x0 (float)
        Left longitude corner to read.
    x1 (float)
        Right longitude corner to read.
    y0 (float)
        Bottom latitude corner to read.
    y1 (float)
        Top latitude corner to read.

    Returns
    -------
    dset (xr.Dataset)
        Merged transport dataset.

    """
    dset = xr.open_mfdataset(
        filenames, concat_dim="nc", combine="nested", chunks=chunks
    )
    ix0, ix1, iy0, iy1 = indices(dset.lon_v, dset.lat_u, x0, x1, y0, y1)
    dset = dset.isel(nx=slice(ix0, ix1), ny=slice(iy0, iy1)).chunk(chunks)
    dset["uRe"] = dset.uRe * 1e-4
    dset["uIm"] = dset.uIm * 1e-4
    dset["vRe"] = dset.vRe * 1e-4
    dset["vIm"] = dset.vIm * 1e-4
    return dset.rename({"uRe": "URe", "uIm": "UIm", "vRe": "VRe", "vIm": "VIm"})



"""Core tools for tidal analysis and prediction."""
from pathlib import Path
import yaml
import numpy as np
import xarray as xr

from oceantide.constituents import V0U


HERE = Path(__file__).parent


def nodal(time: np.ndarray, con: np.ndarray):
    """Nodal correction.

    Parameters
    ----------
    time (1darray)
        Time given as the number of days since 01 Jan 1992 + 48622, or equivalently
        the number of days since 17 Nov 1858.
    con (1darray)
        Constituents to compute.

    Returns
    -------
    pu (1darray)
        Nodal correction pu.
    pf (1darray)
        Nodal correction pf.
    v0u (1darray)
        Nodal correction v0u.

    """
    rad = np.pi / 180.0
    s, h, p, omega = astrol(time)
    sinn = np.sin(omega * rad)
    cosn = np.cos(omega * rad)
    sin2n = np.sin(2 * omega * rad)
    cos2n = np.cos(2 * omega * rad)
    sin3n = np.sin(3 * omega * rad)

    ndict = {
        "M2": {
            "f": np.sqrt(
                (1.0 - 0.03731 * cosn + 0.00052 * cos2n) ** 2
                + (0.03731 * sinn - 0.00052 * sin2n) ** 2
            ),
            "u": np.arctan(
                (-0.03731 * sinn + 0.00052 * sin2n)
                / (1.0 - 0.03731 * cosn + 0.00052 * cos2n)
            )
            / rad,
        },
        "S2": {"f": 1.0, "u": 0.0},
        "K1": {
            "f": np.sqrt(
                (1.0 + 0.1158 * cosn - 0.0029 * cos2n) ** 2
                + (0.01554 * sinn - 0.0029 * sin2n) ** 2
            ),
            "u": np.arctan(
                (-0.1554 * sinn + 0.0029 * sin2n)
                / (1.0 + 0.1158 * cosn - 0.0029 * cos2n)
            )
            / rad,
        },
        "O1": {
            "f": np.sqrt(
                (1.0 + 0.189 * cosn - 0.0058 * cos2n) ** 2
                + (0.189 * sinn - 0.0058 * sin2n) ** 2
            ),
            "u": 10.8 * sinn - 1.3 * sin2n + 0.2 * sin3n,
        },
        "N2": {
            "f": np.sqrt(
                (1.0 - 0.03731 * cosn + 0.00052 * cos2n) ** 2
                + (0.03731 * sinn - 0.00052 * sin2n) ** 2
            ),
            "u": np.arctan(
                (-0.03731 * sinn + 0.00052 * sin2n)
                / (1.0 - 0.03731 * cosn + 0.00052 * cos2n)
            )
            / rad,
        },
        "P1": {"f": 1.0, "u": 0.0},
        "K2": {
            "f": np.sqrt(
                (1.0 + 0.2852 * cosn + 0.0324 * cos2n) ** 2
                + (0.3108 * sinn + 0.0324 * sin2n) ** 2
            ),
            "u": np.arctan(
                -(0.3108 * sinn + 0.0324 * sin2n)
                / (1.0 + 0.2852 * cosn + 0.0324 * cos2n)
            )
            / rad,
        },
        "Q1": {
            "f": np.sqrt((1.0 + 0.188 * cosn) ** 2 + (0.188 * sinn) ** 2),
            "u": np.arctan(0.189 * sinn / (1.0 + 0.189 * cosn)) / rad,
        },
    }

    ncon = len(con)
    pu = np.zeros(ncon)
    pf = np.ones(ncon)
    v0u = np.zeros(ncon)

    for ii, vv in enumerate(con):
        if vv in ndict:
            pu[ii] = ndict[vv]["u"] * rad
            pf[ii] = ndict[vv]["f"]
        if vv in V0U.keys():
            v0u[ii] = V0U[vv]

    return pu, pf, v0u


def astrol(time: np.ndarray):
    """Mean astronomical longitudes  s, h, p, N.

    Parameters
    ----------
    time (1darray)
        Time given as the number of days since 01 Jan 1992 + 48622, or equivalently
        the number of days since 17 Nov 1858.

    Returns
    -------
    s (1darray)
        The mean longitude of moon.
    h (1darray)
        The mean longitude of sun.
    p (1darray)
        The mean longitude of lunar perigee.
    N (1darray)
        The mean longitude of ascending lunar node.

    """
    T = time - 51544.4993
    s = (218.3164 + 13.17639648 * T) % 360
    h = (280.4661 + 0.98564736 * T) % 360
    p = (83.3535 + 0.11140353 * T) % 360
    N = (125.0445 - 0.05295377 * T) % 360
    return s, h, p, N


def arakawa_grid(
    nx: int, ny: int, x0: float, x1: float, y0: float, y1: float, variable: str
):
    """Arakawa grid coordinates for variable.

    Parameters
    ----------
    nx (int)
        Number of grid point along the x direction.
    ny (int)
        Number of grid point along the y direction.
    x0 (float)
        Left limit (degrees).
    x1 (float)
        Right limit (degrees).
    y0 (float)
        Bottom limit (degrees).
    y1 (float)
        Left limit (degrees).
    variable (str)
        Model variable, one of 'h', 'u', 'v'.

    Returns
    -------
    lon (1darray)
        Longitude coordinates in Arakawa grid for variable.
    lat (1darray)
        Latitude coordinates in Arakawa for variable.

    """
    dx = (x1 - x0) / nx
    dy = (y1 - y0) / ny
    xcorner = np.clip(np.arange(x0, x1, dx), x0, x1)[0:nx]
    ycorner = np.clip(np.arange(y0, y1, dy), y0, y1)[0:ny]
    if variable == "h":
        lon = xcorner + dx / 2
        lat = ycorner + dy / 2
    elif variable == "u":
        lon = xcorner
        lat = ycorner + dy / 2
    elif variable == "v":
        lon = xcorner + dx / 2
        lat = ycorner
    else:
        raise ValueError(f"'variable' must be one of 'h', 'u', 'v', got {variable}")
    return lon, lat


def set_attributes(dset: xr.Dataset, dataset_type: str):
    """Set variable attributes for variables from given dataset type.

    Parameters
    ----------
    dset (Dataset)
        Dataset to set attributes from.docs
    dataset_type (str)
        Dataset type, e.g., 'otis'.

    """
    all_attrs = yaml.load(open(HERE / "attributes.yml"), Loader=yaml.Loader)
    if isinstance(dset, xr.Dataset):
        for varname, darr in dset.variables.items():
            darr.attrs.update(all_attrs.get(dataset_type).get(varname, {}))
    elif isinstance(dset, xr.DataArray):
        dset.attrs.update(all_attrs.get(dataset_type).get(dset.name, {}))


def compute_scale_and_offset(vmin: float, vmax: float, nbit: int = 16):
    """Returns the scale_factor and add_offset for packing data.

    Parameters
    ----------
    vmin (float)
        The minumum value in the array to pack.
    vmax (float)
        The maximum value in the array to pack.
    nbit (int)
        The number of bits into which you wish to pack.

    Returns
    -------
    scale_factor (float)
        The scale factor value to use for packing array.
    add_offset (float)
        The add_offset value to use for packing array.

    """
    if vmax == vmin:
        scale_factor = 1
    else:
        scale_factor = (vmax - vmin) / (2**nbit - 1)
    add_offset = vmin + 2 ** (nbit - 1) * scale_factor
    return scale_factor, add_offset



"""Read oceantide file format."""
import xarray as xr

from oceantide.tide import Tide
from oceantide.core.utils import set_attributes


def read_oceantide(
    filename: str = None, engine: str = None, chunks: dict = {}, **kwargs
) -> xr.Dataset:
    """Read oceantide file format.

    The oceantide format has complex constituents variables split into real and imag
    variables allowing for better packing and making it easier to support netcdf.

    parameters
    ----------
    filename (str)
        Name of Oceantide dataset to read.
    engine (str)
        Engine to pass to xr.open_dataset, guessed by xarray based on filename
        extension if not specified.
    chunks (dict)
        Chunk sizes along each dimensions to pass to xr.open_dataset, by default use
        chunking specified on disk.
    kwargs
        Extra kwargs to pass to xr.open_dataset.

    Returns
    -------
    dset (xr.Dataset)
        Formatted dataset with complex variables and the tide accessor.

    """
    dset = xr.open_dataset(filename, engine=engine, chunks=chunks, **kwargs)
    dsout = dset[["dep"]]
    for v in ["h", "u", "v"]:
        dsout[v] = dset[f"{v}_real"] + 1j * dset[f"{v}_imag"]
    set_attributes(dsout, "dataset")
    dsout["con"] = dsout.con.astype("U4")
    return dsout



"""Tide xarray accessor."""
import os
import re
import glob
import logging
from pathlib import Path
from importlib import import_module
from inspect import getmembers, isfunction
import datetime
import numpy as np
import dask.array as da
import pandas as pd
import xarray as xr

from oceantide.core.utils import nodal, set_attributes
from oceantide.constituents import OMEGA


logger = logging.getLogger(__name__)

HERE = Path(__file__).parent


class Plugin(type):
    """Add output functions as bound methods at class creation."""

    def __new__(cls, name, bases, dct):
        module_names = [
            f.replace(".py", "")
            for f in glob.glob1(os.path.join(HERE, "output"), "*.py")
        ]
        modules = [import_module(f"oceantide.output.{name}") for name in module_names]
        for module_name, module in zip(module_names, modules):
            for func_name, func in getmembers(module, isfunction):
                if func_name == f"to_{module_name}":
                    dct[func_name] = func
        return type.__new__(cls, name, bases, dct)


@xr.register_dataset_accessor("tide")
class Tide(metaclass=Plugin):
    """Xarray accessor object to process tide constituents dataset."""

    def __init__(self, xarray_obj):
        self._obj = xarray_obj
        self._validate()

    def __repr__(self):
        return re.sub(r"<.+>", f"<{self.__class__.__name__}>", str(self._obj))

    def _set_attributes_output(self, dset):
        """Set attributes in output timeseries dataset."""
        set_attributes(dset, "timeseries")
        dset.attrs = {"description": "Tide elevation and currents time series"}
        return dset

    def _validate(self):
        """Check if dataset object has already been constructed.

        The following are checked to ensure it has appropriate format:
            * Required complex variables are present and share same coordinates.
            * Constituents coordinate appropriately formatted.

        """
        required_vars = ["h", "u", "v"]
        if not set(required_vars).issubset(self._obj.data_vars):
            logger.debug(
                f"Tide accessor requires variables {required_vars} in dataset for full"
                f" functionality but only found {list(self._obj.data_vars.keys())}."
            )

        for v in required_vars:
            if v in self._obj.data_vars and not np.iscomplexobj(self._obj[v]):
                raise ValueError(f"Variable {v} must be complex type.")

        vars_dims = [self._obj[v].dims for v in required_vars if v in self._obj]
        if len(set(vars_dims)) != 1:
            raise ValueError(f"Variables {required_vars} must share a common grid.")

        if self._obj.con.dtype.kind != "U":
            raise ValueError(
                f"`con` must be Unicode string dtype, found {self._obj.con.dtype}"
            )

    def amplitude(self, component="h"):
        """Tidal amplitude.

        :math:`\\lambda=\\sqrt{\\Re(z)+\\Im(z)}`

        Args:
            - component (str): Tidal component to calculate amplitude from,
              one of 'h', 'u', 'v'.

        Returns:
            - amp (DataArray): Amplitudes for component :math:`\\lambda(con,lat,lon)`.

        """
        darr = np.absolute(self._obj[component])
        darr.name = f"amp{component}"
        return darr

    def phase(self, component="h"):
        """Tidal phase relative to GMT.

        Args:
            - component (str): Tidal component to calculate amplitude from,
              one of 'h', 'u', 'v'.

        Returns:
            - phi (DataArray): Phases for component :math:`\\phi(con,lat,lon)`.

        """
        darr = self._obj[component]
        darr = np.rad2deg(np.arctan2(-darr.imag, darr.real)) % 360
        darr.name = f"phi{component}"
        set_attributes(darr, "dataset")
        return darr

    def predict(self, times, time_chunk=50, components=["h", "u", "v"]):
        """Predict tide timeseries.

        Args:
            times (arr): Array of datetime objects or DataArray of times to predict tide over. If an array, a new times dimension will be created.
            time_chunk (float): Time chunk size so that computation fit into memory.
            components (list): Tide variables to predict.

        Returns:
            ds (Dataset): Predicted tide timeseries components :math:`\\eta(time,lat,lon)`.

        """
        if not components or {"h", "u", "v"} - set(components) == {"h", "u", "v"}:
            raise ValueError("Choose at least one tide component h, u or v to predict")

        if isinstance(times, datetime.datetime):
            times = [times]
        if isinstance(times, (list, tuple, np.ndarray, pd.DatetimeIndex)):
            seconds_array = pd.array(times).view(int) / 1e9 - 694224000
            tsec = xr.DataArray(
                data=seconds_array,
                coords={"time": times},
                dims=("time",),
            ).chunk({"time": time_chunk})
        elif isinstance(times, xr.DataArray):
            tsec = times.astype(int) / 1e9 - 694224000
        else:
            raise TypeError(
                "times argument must be a list of datetimes, pandas.DatetimeIndex, "
                "numpy array of datetime64 or xarray DataArray of datetime64"
            )

        cons = list(self._obj.con.values)

        # Slice out unsupported cons
        non_supported = set(cons) - OMEGA.keys()
        if non_supported:
            logger.warning(f"Cons {non_supported} not supported and will be ignored")
            cons = [con for con in cons if con not in non_supported]
            ds = self._obj.sel(con=cons)
        else:
            ds = self._obj

        pu, pf, v0u = nodal(tsec[0] / 86400 + 48622.0, cons)

        # Variables for calculations
        pf = xr.DataArray(pf, coords={"con": cons}, dims=("con",))
        pu = xr.DataArray(pu, coords={"con": cons}, dims=("con",))
        v0u = xr.DataArray(v0u, coords={"con": cons}, dims=("con",))
        omega = xr.DataArray(
            data=[OMEGA[c] for c in cons], coords={"con": cons}, dims=("con",)
        )

        cos = da.cos(tsec * omega + v0u + pu)
        sin = da.sin(tsec * omega + v0u + pu)

        dset = xr.Dataset()
        if "dep" in ds.data_vars:
            dset["dep"] = ds.dep
        if "h" in components:
            dset["h"] = cos * pf * ds["h"].real - sin * pf * ds["h"].imag
        if "u" in components:
            dset["u"] = cos * pf * ds["u"].real - sin * pf * ds["u"].imag
        if "v" in components:
            dset["v"] = cos * pf * ds["v"].real - sin * pf * ds["v"].imag
        dset = dset.sum(dim="con", skipna=False)
        dset = self._set_attributes_output(dset)

        return dset
